unit modZ80;

interface
{$i tables.inc}

procedure Z80Reset;
procedure execute(var local_tstates: integer);

procedure pop(var wpval:Word);
procedure push(wpval:Word);

function peekb(addr : integer) : Byte;
procedure pokeb(addr : integer; val : Byte);
procedure pokew(addr : integer; val : Word);
function peekw(Addr:Word):Word;

procedure Z80_NMI;
procedure create_cpu;

implementation

uses Windows, Forms, Sysutils, mmsystem, UMain,
     modMain, modWaveout, modAY8912,modTape, ModDisplay;

function peekb(addr : integer) : Byte;
begin
    result := gRAMPage[glPageAt[(addr and $C000) shr 14], (addr And $3FFF)];
end;

function peekw(Addr:Word):Word;
begin
 peekw:=(peekb(addr+1) shl 8) or peekb(addr);
end;

procedure pokeb(addr : integer; val : Byte);
begin
    If (addr < 16384) and (glPageAt[0]<>nMF128start) Then Exit;// ROM
    gRAMPage[glPageAt[(addr and $C000) shr 14], addr and $3FFF] := val;
end;

procedure pokew(addr : integer; val:word);
var value:pair absolute val;
begin
    pokeb(addr, value.l); pokeb(addr + 1, value.h);
end;

var t_state:integer;

{$i procs.inc}
{$i cbops.inc}
{$i edops.inc}
{$i z80ops.inc}

function specinterrupt : integer;
    var lSleep : integer;
{$IFDEF USE_SOUND}
        lCounter: integer;
        PDst : Pointer;
        PSrc : Pointer;
{$ENDIF}
begin
    interruptCounter := interruptCounter + 1;

    // Attribute flash (state alters every 1/2 second)
    If (interruptCounter Mod 25) = 0 Then
            display.Flash_Status:=not display.Flash_Status;

    screenPaint;

    // PERFORM Z80 hardware functions
     halt:=false;

    // If it's a maskable interrupt
    If IFF1 = 0 Then
        specinterrupt := 0
    Else
    begin
     iff2:=iff1;
     iff1:=0;
     halt:=false;
     if peekb(pc.w)=$76 then inc(pc.w);
     push(pc.w);
     if im=2 then
      Begin
       pc.W:=peekw((ir.h shl 8) or 255);
       inc(ir.l);
       specinterrupt := 19;
      end
      else
      Begin
       inc(ir.l);
       pc.w:=$38; {Z80[255];}
       specinterrupt := 13;
      end;
    End;

    {$IFDEF USE_SOUND}
    If gbSoundEnabled Then
    begin
        inc(glBufNum);

        For lCounter := glWavePtr to WAV_BUFFER_SIZE do
            gcWaveOut[lCounter] := gcWaveOut[glWavePtr - 1];

        PDst := gtWavHdr[glBufNum].lpData;

        pSrc := @gcWaveOut[0];
        CopyMemory(PDst, PSrc, WAV_BUFFER_SIZE);

        waveOutWrite(glphWaveOut,@gtWavHdr[glBufNum], sizeof(WAVEHDR));

        If glBufNum = NUM_WAV_BUFFERS Then glBufNum := 0;
    End;
    glWavePtr := 0;
    {$ENDIF}

    //Keep the emulation running at the correct speed by
    // adding a delay to ensure that interrupts are
    // generated at the correct frequency
    lSleep := glInterruptTimer - integer(timeGetTime()) + glDelayOverage;
    If lSleep < 0 Then
    begin
        If glDelayOverage < -40 Then
            glDelayOverage := -40
        Else
            glDelayOverage := lSleep;
    end;

    If lSleep > 0 Then
    begin
        Sleep(lSleep);
        glDelayOverage := glDelayOverage + (glInterruptDelay - lSleep);
        If glDelayOverage > 0 Then glDelayOverage := 0;
    End;

    glInterruptTimer := integer(timeGetTime()) + glInterruptDelay;
end;


function nxtpcb: integer;
  var dummy: integer;
begin
   inc(ir.l);
   dummy:=peekb(pc.W);
   {$IFDEF USE_DEBUGGING}
    ReportDebug(0, 'next PC', pc.w, false);
   {$ENDIF}
   inc(pc.w);
   result := dummy;
end;

procedure execute(var local_tstates: integer);
    var
      intCtrTemp : integer;
      d : integer; lTemp : integer;
begin

    //Execute one interrupt duration
    intCtrTemp := interruptCounter;

    repeat
      If (local_tstates >= 0) Then  // Trigger an interrupt
      begin
       {$IFDEF USE_DEBUGGING}
         ReportDebug(0,'Interrupt', pc.w, true);
       {$ENDIF}

        local_tstates := local_tstates - glTstatesPerInterrupt - specinterrupt;
      end;

     // Hook the TRDOS
     if glEmulatedTRDOS then
      begin                             //nromstart+3=TRDOS
        if (PC.W>=$4000) and (glPageAt[0]=nromstart+3) then
         glPageAt[0]:=byte(glLastOut7ffd and 16>0)+nRomStart
        else
        if (PC.W and $ff00=$3d00) and (glPageAt[0]=nRomStart+1) then
         glPageAt[0]:=nRomStart+3; // TRDOS
       end;

      t_state:=0;EvalBreakPoints;
      if halt then z80[0] else z80[nxtpcb];
      inc(local_tstates,t_state);

      {$IFDEF USE_SOUND}
       AddSoundWave(T_STATE);
      {$ENDIF}
      Paint_Next_Line(T_STATE);
    until (intCtrTemp <> interruptCounter);
end;


procedure Z80_NMI;
var temp:byte;
begin
  iff2:=iff1;
  iff1:=0;
  inc(ir.l);
  push(pc.W);
  pc.W:=102;
  inc(t_state,11);
  if glEmulatedModel in [3,4] {Plus3} then inb(63) else inb(191);
 end;


procedure Z80Reset;
begin
	iff1:=0;iff2:=0;
        bit7_r:=0;
        halt:=false;
        af.W:=0;af2.W:=0;
	hl.W:=0;hl2.W:=0;
        de.W:=0;de2.W:=0;
        bc.W:=0;bc2.W:=0;
	ix.W:=$FFFF;iy.W:=$FFFF;

        im:=1;ir.W:=0;
        sp.W:=0;pc.W:=0;
        glLastOut7FFD:=0;
        glLastOut1FFD:=0;
        glSoundRegister:=0;
	glvline:=0;
        t_state:=0;
        display.border:=7;


    SetEmulatedModel(strtoint(GetSetting('EmulatedModel', '0')));
    {$IFDEF USE_SOUND}
      if assigned(AY8912) then AY8912.Destroy;
      AY8912:=TAY8912.Create(1773000, WAVE_FREQUENCY, 8);
    {$ENDIF}
end;

procedure calc_intable;
var i,j:Byte;
begin
  for i:=0 to 255 do
  begin
        { $ifndef assembler}
        if i=0 then j:=64 else j:=0;
	add8_table[i]:=(i and 168) or (j); (* Just HVC to do! *)
	sub8_table[i]:=(i and 168) or (j) or 2;
	cpsub8_table[i]:=(i and 128) or (j) or 2; (* 5 bits left.. *)
        { $endif}
        in_table[i]:=(i and 168) or (parity[i]);
	in_table[0]:=in_table[0] or 64; (* Don't forget to keep old C flag *)
  end;
end;

procedure create_cpu;
begin
 createcb;createed;createz80;
 calc_intable;
end;

end.
