unit UVideoRendererRGB;

interface

uses Windows, ActiveX, DirectShow, DirectDraw, SysUtils, UVideoRConst,
     UVideoRendererClass;

type
  TVideoRendererRGB = class(TVideoRendererClass)
  private
    pDDraw: IDirectDraw;
    pPrimary: IDirectDrawSurface;
    pBackBuffer: IDirectDrawSurface;
    pClipper: IDirectDrawClipper;

    // --- window we belong to
    pHWND: HWND;
    pVideoMode: TVideoMode;
  public

    function Stop: HResult;override;
    function GetVideoMode: TVideoMode;override;

    function Draw(buffer: Pointer; rect: TRect; FrameNumber: Integer; InvertFlag: DWord): HResult;override;
    function UpdateFrame(rect: TRect): HResult;override;

    function InitDDraw(fHwnd: HWND; Width, Height: Integer): HResult;override;
    procedure Close;override;
    procedure OnDesktopChange(Sender: TObject; fHwnd:HWND; cX, cY: Integer; bpp: Byte);override;

    procedure ClearToBlack;override;

    constructor Create; override;
    destructor Destroy; override;
  end;

implementation

procedure TVideoRendererRGB.OnDesktopChange(Sender: TObject; fHwnd: HWND; cX, cY: Integer; bpp: Byte);
begin
  if pDDraw <> nil then begin
     InitDDraw(fHwnd, cx, cy);
  end;
end;

function TVideoRendererRGB.UpdateFrame(rect: TRect): HResult;
 var srcRect, destRect: TRect;
     pt: TPoint;
     fx: TDDBLTFX;
     hr: HResult;
begin
  if (pPrimary = nil) or (pBackBuffer = nil) or (pDDraw = nil) then begin result:= DWord(-1); exit; end;
  if pHWND <> 0 then begin
    srcRect.Left:= 0;
    srcRect.Top:= 0;
    srcRect.Right:= pWidth;
    srcRect.Bottom:= pHeight;

    GetClientRect(pHWND, destRect);

    pt.x:= 0; pt.y:= 0;
    ClientToScreen(pHWND, pt);
    OffsetRect(destRect, pt.x, pt.y);

    // --- draw the backbuffer to the front buffer
    Zeromemory(@fx, sizeof(fx));
    fx.dwSize:= sizeof(fx);

    hr:= pPrimary.Blt(@destRect, pBackBuffer, @srcRect, DDBLT_ASYNC or DDBLT_WAIT, @fx);
    if hr = DD_OK then begin result:= NoError; exit; end;

    if hr = DDERR_SURFACELOST then begin
      result:= VRGB_ERROR; exit;
    end;

    if hr <> DDERR_WASSTILLDRAWING then begin
      result:= VRGB_ERROR; exit;
    end;
    result:= NoError;
  end;
end;

procedure TVideoRendererRGB.ClearToBlack;
 var x: TDDBLTFx;
begin
  if (pDDraw <> nil) and (pPrimary <> nil) then begin

    ZeroMemory(@x, sizeof(x));
    x.dwSize:= sizeof(x);
    x.dwFillColor:= 0;

    pPrimary.Blt(nil, nil, nil, DDBLT_COLORFILL or DDBLT_WAIT, @x);
  end;
end;

function TVideoRendererRGB.Draw(buffer: Pointer; rect: TRect; FrameNumber: Integer; InvertFlag: DWord): HResult;
 var hr: HResult;
     desc: TDDSurfaceDesc;
     i: DWord;
     pBDest, pBSrc: PByteArray;
     pCopyCount: DWord;
     pOffsSrc: DWord;
     pOffsDest: DWord;

     srcRect: TRect;
     destRect: TRect;
     pt: TPoint;

     fx: TDDBLTFx;

begin
  result:= VRGB_ERROR;
  if (pDDraw = nil) or (pBackBuffer = nil) or (pPrimary = nil) then exit;

  ZeroMemory(@Desc, sizeof(desc));
  desc.dwSize:= Sizeof(desc);

  // --- Lock the backbuffer surface to get its surface pointer
  hr:= pBackBuffer.Lock(nil, desc, DDLOCK_SURFACEMEMORYPTR or DDLOCK_WRITEONLY or DDLOCK_WAIT, 0);
  if failed(hr) then exit;

  // --- Copy the pixels

  //  prepare all needed numbers so we don't need to do it in the CYCLE
  pBDest:= desc.lpSurface;
  pBSrc:= buffer;
  pCopyCount:= pWidth * pBPP;
  pOffsSrc:= 0;

  case InvertFlag of
    INVERT_NONE : begin
                    pOffsDest:= (desc.dwHeight-1) * desc.lPitch;
                    for i:= 0 to desc.dwHeight-1 do begin
                      Move( pBSrc^[pOffsSrc], pBDest^[pOffsDest], pCopyCount);
                      pOffsSrc:= pOffsSrc + pCopyCount;
                      pOffsDest:= pOffsDest - desc.lPitch;
                    end;
                  end;
    INVERT_UPDOWN: begin
                    pOffsDest:= 0;
                    for i:= 0 to desc.dwHeight-1 do begin
                      Move( pBSrc^[pOffsSrc], pBDest^[pOffsDest], pCopyCount);
                      pOffsSrc:= pOffsSrc + pCopyCount;
                      pOffsDest:= pOffsDest + desc.lPitch;
                    end;
                   end;
  end;

  pBSrc:= nil; pBDest:= nil;
  pBackBuffer.Unlock(nil);

  if pHWND <> 0 then begin
    srcRect.Left:= 0;
    srcRect.Top:= 0;
    srcRect.Right:= pWidth;
    srcRect.Bottom:= pHeight;

    GetClientRect(pHWND, destRect);

    pt.x:= 0; pt.y:= 0;
    ClientToScreen(pHWND, pt);
    OffsetRect(destRect, pt.x, pt.y);

    // --- draw the backbuffer to the front buffer
      Zeromemory(@fx, sizeof(fx));
      fx.dwSize:= sizeof(fx);

      hr:= pPrimary.Blt(@destRect, pBackBuffer, @srcRect, DDBLT_ASYNC or DDBLT_WAIT, @fx);
      if hr = DD_OK then begin result:= NoError; exit; end;

      if hr = DDERR_SURFACELOST then begin
        result:= VRGB_ERROR; exit;
      end;

      if hr <> DDERR_WASSTILLDRAWING then begin
        result:= VRGB_ERROR; exit;
      end;
      result:= NoError;
  end;
end;

function TVideoRendererRGB.Stop: HResult;
begin
  result:= S_OK;
end;

function TVideoRendererRGB.GetVideoMode: TVideoMode;
begin
  result:= pVideoMode;
end;

function TVideoRendererRGB.InitDDraw(fHwnd: HWND; Width, Height: Integer): HResult;
 var hr: HResult;
     desc: TDDSurfaceDesc;
     x: TDDBLTFX;
     sr, dr: TRect;
begin
  if (Width <=0 ) or (Height <=0) then begin result:= VRGB_WRONG_SIZE; exit; end;

  pClipper:= nil;
  pBackBuffer:= nil;
  pPrimary:= nil;
  pDDraw:= nil;

  pHWND:= fHWND;

  hr:= DirectDrawCreate(nil, pDDraw, nil);
  if failed(hr) then begin result:= VRGB_CANNOT_CREATE_DDRAW; exit; end;

  hr:= pDDraw.SetCooperativeLevel(pHWND, DDSCL_NORMAL);
  if faileD(hr) then begin Close; result:= VRGB_ERROR; exit; end;

  // --- Create the Primary Surface
  ZeroMemory(@desc, sizeof(desc));
  desc.dwSize:= Sizeof(Desc);
  desc.dwFlags:= DDSD_CAPS;
  desc.ddsCaps.dwCaps:= DDSCAPS_PRIMARYSURFACE or DDSCAPS_VIDEOMEMORY;

  hr:= pDDraw.CreateSurface(desc, pPrimary, nil);
  if faileD(hr) then begin Close; result:= VRGB_ERROR; exit; end;

  // --- Make a Clipper for it
  hr:= pDDraw.CreateClipper(0, pClipper, nil);
  if failed(hr) then begin Close; result:= VRGB_ERROR; exit; end;

  hr:= pClipper.SetHWnd(0, pHWND);
  if failed(hr) then begin Close; result:= VRGB_ERROR; exit; end;

  // --- Assign the Clipper to the Primary Surface
  hr:= pPrimary.SetClipper(pclipper);
  if failed(hr) then begin Close; result:= VRGB_ERROR; exit; end;


  // --- Create the BackBuffer
  ZeroMemory(@Desc, sizeof(desc));
  Desc.dwSize:= Sizeof(desc);
  desc.dwFlags:= DDSD_CAPS or DDSD_HEIGHT or DDSD_WIDTH;
  desc.ddsCaps.dwCaps:= DDSCAPS_OFFSCREENPLAIN;
  desc.dwWidth:= Width;
  desc.dwHeight:= Height;
  hr:= pDDraw.CreateSurface(desc, pBackBuffer, nil);
  if failed(hr) then begin Close; result:= VRGB_ERROR; exit; end;

  pWidth:= Width;
  pHeight:= Height;

  hr:= pBackBuffer.Lock(nil, desc, DDLOCK_SURFACEMEMORYPTR or DDLOCK_WRITEONLY or DDLOCK_WAIT, 0);
  if failed(hr) then begin result:= VRGB_ERROR; exit; end;
  pBackBuffer.Unlock(0);

  case desc.ddpfPixelFormat.dwRGBBitCount of
    8 : result:= VRGB_ERROR;
    16 : begin
           pBPP:= 2;
           pVideoMode:= VIDEOMODE_RGB16;
         end;
    24 : begin
           pBPP:= 3;
           pVideoMode:= VIDEOMODE_RGB24;
         end;
    32 : begin
           pBPP:= 4;
           pVideoMode:= VIDEOMODE_RGB32;
         end;
  end;
  result:= NoError;

  ZeroMemory(@x, sizeof(x));
  x.dwSize:= sizeof(x);
  x.dwFillColor:= 0;
  sr.left:= 0; sr.top:= 0; sr.right:= pWidth; sr.Bottom:= pHeight;
  dr:= sr;
  pBackBuffer.Blt(@dr, nil, @sr, DDBLT_WAIT, @X);
  pPrimary.Blt(nil, nil, nil, DDBLT_WAIT, @X);
end;

procedure TVideoRendererRGB.Close;
begin
  if pDDraw <> nil then begin
                          pDDraw.RestoreDisplayMode;
                          pDDraw.SetCooperativeLevel(pHWND, DDSCL_NORMAL);
                        end;
  pClipper:= nil;
  pBackBuffer:= nil;
  pPrimary:= nil;
  pDDraw:= nil;

  pVideoMode:= VIDEOMODE_NONE;
  pBPP:= 0;
end;

destructor TVideoRendererRGB.Destroy;
begin
  Close;
  inherited;
end;

constructor TVideoRendererRGB.Create;
begin
  inherited;

  pVideoMode:= VIDEOMODE_NONE;
  pBPP:= 0;

  pDDraW:= nil;
  pPrimary:= nil;
  pBackBuffer:= nil;
  pClipper:= nil;
end;


end.
