unit modMain;

{' /*******************************************************************************
'   modMain.pas within DelphiSpec.dpr
'
'   Public variable declarations, startup and initialization code
'
'   Original author (vbSpec): Chris Cowley <ccowley@grok.co.uk>
'   Copyright (C)1999-2000 Grok Developments Ltd.
'   http://www.grok.co.uk/
'
'   Translation to Delphi Object Pascal by
'   Jari Korhonen <jarit.korhonen@luukku.com>
'   Copyright (C) 2001- Jari Korhonen
'
'   This program is free software; you can redistribute it and/or
'   modify it under the terms of the GNU General Public License
'   as published by the Free Software Foundation; either version 2
'   of the License, or (at your option) any later version.
'   This program is distributed in the hope that it will be useful,
'   but WITHOUT ANY WARRANTY; without even the implied warranty of
'   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
'   GNU General Public License for more details.
'
'   You should have received a copy of the GNU General Public License
'   along with this program; if not, write to the Free Software
'   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
'
' *******************************************************************************/}


interface

uses Windows, Graphics, Mmsystem, IniFiles,
     Sysutils, Forms, Extctrls, Math,classes;

type
   pair=record
        case boolean of
           false: (l,h:Byte);
           true: (W:Word);
        end;
    quadruple=record
           case byte of
              0: (b1,b2,b3,b4:Byte);
              1: (w1,w2:Word);
              2: (q:longint);
           end;

// Flag bit positions
Const
  BORDERWIDTH = 20;
  Z80RUNNING=1;       //Normally
  Z80PAUSED=2;        //When in setting screens
  Z80STOPPED=3;       //When losing focus

// Screen things
  nPixelsWide = 256;
  nPixelsHigh = 192;
  nCharsWide = 32;
  nCharsHigh = 24;
  FirstAttr = nPixelsHigh * nCharsWide;
  //lastAttr = FirstAttr + (nCharsHigh * nCharsWide);
  nRomStart=8+7; // First Rom in Data
  nMF128start=nromstart+4; // MF128 Rom

var
glTstatesPerInterrupt: integer;
glTstatesPerLine: integer;
Parity: array[0..256] of Boolean;
glInterruptTimer: integer;

gRAMPage: array[0..19, 0..16383] of Byte; //(pages 0 to 7 are RAM, and pages 8 to 11 are ROM)
glPageAt: array[0..4] of integer;
glLastOut7FFD: integer;
glLastOut1FFD: integer;
glUseScreen: integer;
glEmulatedModel: integer;
glEmulatedTRDOS:Boolean;
glEmulatedMF128:Boolean;
glEmulatedMF3: Boolean;
glEmulatedFlash: Boolean;
glAutoReboot: Boolean;
glResolution: Integer; // Resolutionsetting
glBetaDisk: Boolean; // Allows Betadisk in each Machine

Stickvalue: byte;

{$IFDEF USE_SOUND}
gbSoundEnabled: Boolean;
gbSpeakerEnabled: Boolean;
gbAy8912Enabled: Boolean;
glSoundRegister: integer;       // Contains currently indexed AY-3-8912 sound register
glBufNum: integer;              // ID of the last Wave buffer used
{$ENDIF}

// Main Z80 registers

    af, bc, de, hl:pair;
    ix, iy:pair;
    sp, pc:pair;
    ir: Pair;
    rtemp,im   : Byte;
    AF2,BC2,DE2,HL2: Pair; { == AF' BC' DE' HL'}
    iff1, iff2, bit7_r: Byte;
    halt:boolean;
    qtemp:Quadruple;
    ptemp:pair;

GLVLine:integer=0;
GLJoystick:record up,down,left,right,fire:TShortCut end;

// Addressable Memory (16KB ROM at 0x0000, 48KB RAM at 0x4000)
// We also place a shadow copy of the ROM above the addressable memory space
// at 0x10000 so we don't have to check for regPC > 65535 at each instruction
mem: array[0..81919] of Byte;


// Interrupts/Screen refreshing
interruptCounter: integer;
glInterruptDelay: integer;
glDelayOverage: integer;

// Keypresses
keyB_SPC: integer;
keyH_ENT: integer;
keyY_P: integer;
key6_0: integer;
key1_5: integer;
keyQ_T: integer;
keyA_G: integer;
keyCAPS_V: integer;


glDisplayHeight: integer;
glDisplayWidth: integer;
glRealBorder: Boolean;
glDisplayMode: integer;
glFrameSkip: Byte;

{$IFDEF USE_SOUND}
glBeeperVal: integer;
{$ENDIF}

function InitializeWaveOut : Boolean;
function GetSetting(settingName: string; defaultValue:string): string;
procedure SaveSetting(settingName: string; value:string);

procedure SetDisplay(Mode:Byte;Border,Flash:Boolean;Skip:byte);
procedure SetDisplaySize(lWidth : integer; lHeight : integer);
procedure SetKempston(Up,Down,Left,Right,Fire: TShortCut);
procedure SetEmulatedModel(lModel : integer);
procedure SetStatus(sMsg : String);
procedure Main;
procedure CloseWaveOut;
procedure resetKeyboard;
procedure emptyRAMs;
function ramCrc(page: integer): integer;
function GetAppPath: string;

var
 glNewBorder : integer;
 inb:function(port : Word) : byte;
 outb:procedure(port : Word; outbyte : byte);

procedure Paint_Next_Line(Line:Integer);
procedure screenPaint;
procedure SpeakerOut(Value: byte);

implementation

uses UMain, USetup, modAY8912, ModTape, ModBetadisk,
     modWaveOut, modZ80, ModDisplay, ModLoader,
     ModZX48,ModZX128,
     ModPlus2,
     ModPlus3,ModPentagon,ModScorpion;

const t_state:integer=0;
      B_SPEAKER = $10;
      B_MIC     = $08;

procedure SpeakerOut (value:byte);
var spk,mic:Boolean;
begin
{$IFNDEF USE_SOUND} exit; {$ENDIF}

 spk:=(value and B_SPEAKER=B_SPEAKER);
 mic:=(value and B_MIC=B_MIC);

 if spk then glbeeperVal:=31 else
 if mic then glbeeperVal:=2 else glbeeperVal:=0;
{$IFDEF USE_DEBUGGING}
 if glbeeperval>0 then ReportDebug(0,'glBeeperVal=on',0,True);
{$ENDIF}

end;

procedure Paint_Next_Line(Line:Integer);
begin
   inc(t_state,line);
   if t_state>=glTstatesPerLine then
   with display do
   begin
    glvline:=(glvline+1) mod (192+top shl 1);t_state:=0;
    Updateline(glvline-top);
   end;
end;

procedure screenPaint;
var i:integer;
begin
 with frmMainWnd do
 begin
  //for i:=-top to 191+top do display.Updateline(i);
  //for i:=0 to 191 do display.Updateline(i);
  display.repaint(frmmainwnd.clientrect);
 end;
end;

function ramCrc(page: integer): integer;
  var Crc, ii : integer;
begin
  Crc := 0;
  for ii := 0 to 16383 do
    Crc := Crc + gRAMPage[page, ii];
  ramCRC := Crc;
end;

procedure emptyRAMs;
  var ii, jj: integer;
begin
  for ii := 0 to 11 do
    for jj := 0 to 16383 do
      gRAMPage[ii, jj] := 0;
end;



//Function emulates VB's App.Path
function GetAppPath: string;
begin
  GetAppPath := ExtractFilePath(Application.Exename);
end;


//Function emulates VB's GetSetting API with TIniFile
function GetSetting(settingName: string; defaultValue:string): string;
  var IniFile: TIniFile;
begin
  IniFile := TIniFile.Create(GetAppPath+'\YASE.INI');
  GetSetting := IniFile.ReadString('YaSE', settingName, defaultvalue);
  IniFile.Free;
end;

//Function emulates VB's SaveSetting API with TIniFile
procedure SaveSetting(settingName: string; value:string);
  var IniFile: TIniFile;
begin
  IniFile := TIniFile.Create(GetAppPath+'\YASE.INI');
  IniFile.WriteString('YaSE', settingName, value);
  IniFile.Free;
end;


procedure CloseWaveOut;
{$IFDEF USE_SOUND}
  var lRet : integer;
{$ENDIF}
begin
    {$IFDEF USE_SOUND}
        waveOutReset(glphWaveOut);
        For lRet := 1 To NUM_WAV_BUFFERS do
            waveOutUnprepareHeader(glphWaveOut,
              @gtWavHdr[lRet], sizeof(WAVEHDR));

        For lRet := 1 To NUM_WAV_BUFFERS do
        begin
            GlobalUnlock(ghMem[lRet]);
            GlobalFree(ghMem[lRet]);
        end;

        waveOutClose(glphWaveOut);

        gbSoundEnabled := False;
    {$ENDIF}
End;


Function InitializeWaveOut : Boolean;
{$IFDEF USE_SOUND}
  var lRet : integer; sErrMsg : String; lCounter : integer; lCounter2 : integer;
{$ENDIF}
begin
    {$IFDEF USE_SOUND}
        If StrToInt(GetSetting('SoundEnabled', '1')) = 0 Then
        begin
            InitializeWaveOut := False;
            Exit;
        End;

        With gtWavFormat do
        begin
            wFormatTag := WAVE_FORMAT_PCM;
            nChannels := 1;
            nSamplesPerSec := WAVE_FREQUENCY;
            nAvgBytesPerSec := WAVE_FREQUENCY;
            nBlockAlign := 1;
            wBitsPerSample := 8;
            cbSize := 0;
        End;

        lRet := waveOutOpen(@glphWaveOut, WAVE_MAPPER, @gtWavFormat, 0, 0, CALLBACK_NULL);
        If lRet <> MMSYSERR_NOERROR Then
        begin
            SetLength(sErrMsg,  256);
            waveOutGetErrorText(lRet, PChar(sErrMsg), 255);
            Application.MessageBox(
              PChar('Error initialising WaveOut device.' + #13#10#13#10 + sErrMsg),
              nil,
              MB_OK or MB_ICONEXCLAMATION);
            InitializeWaveOut := False;
            Exit;
        End;

        For lCounter := 0 To NUM_WAV_BUFFERS do
        begin
            ghMem[lCounter] := GlobalAlloc(GPTR, WAV_BUFFER_SIZE+100);
            gpMem[lCounter] := GlobalLock(ghMem[lCounter]);
            gtWavHdr[lCounter].lpData := gpMem[lCounter];
            gtWavHdr[lCounter].dwBufferLength := WAV_BUFFER_SIZE;
            gtWavHdr[lCounter].dwUser := 0;
            gtWavHdr[lCounter].dwFlags := 0;
            gtWavHdr[lCounter].dwLoops := 0;
            gtWavHdr[lCounter].lpNext := nil;

            lRet := waveOutPrepareHeader(glphWaveOut,
              @gtWavHdr[lCounter],
              sizeof(WAVEHDR));

            If lRet <> MMSYSERR_NOERROR Then
            begin
                SetLength(sErrMsg,  256);
                waveOutGetErrorText(lRet, PChar(sErrMsg), 255);
                Application.MessageBox(
                  PChar('Error preparing wave header.' + #13#10#13#10 + sErrMsg),
                  nil,
                  MB_OK or MB_ICONEXCLAMATION);
                waveOutClose(glphWaveOut);
                For lCounter2 := 0 To NUM_WAV_BUFFERS do
                begin
                    GlobalUnlock(ghMem[lCounter2]);
                    GlobalFree(ghMem[lCounter2]);
                end;
                InitializeWaveOut := False;
                Exit;
            End;
        end;
    {$ENDIF}
        InitializeWaveOut := True;
End;


procedure initParity;
    var lCounter : integer; j : Byte; p : Boolean;
    var powval: integer;
begin
    For lCounter := 0 To 255 do
    begin
        p := True;
        For j := 0 To 7 do
        begin
            PowVal := Trunc(IntPower(2, j));
            If (lCounter And powVal) <> 0 Then p := Not p;
        end;
        Parity[lCounter] := p;
    end;
End;


procedure Main; // Init Display here
  var ii: integer;
begin
    Create_CPU;
    emptyRAMs;
    glDisplayWidth := StrToInt(GetSetting('DisplayWidth', '256'));
    glDisplayHeight := StrToInt(GetSetting('DisplayHeight', '192'));
    glDisplayMode := StrToInt(GetSetting('DisplayMode','0'));
    glRealBorder:= StrToInt(GetSetting('DisplayBorder','1'))=1;
    glFrameSkip:=StrToInt(GetSetting('DisplayFrameSkip','0'));
    glEmulatedFlash:=StrToInt(GetSetting('EmulatedFlash', '1'))=1;

  with glJoystick do
  begin
   up:= StrToInt(GetSetting('VK_UP',   inttostr(ord('Q'))));
   down:= StrToInt(GetSetting('VK_DOWN',inttostr(ord('A'))));
   left:= StrToInt(GetSetting('VK_LEFT',inttostr(ord('O'))));
   right:= StrToInt(GetSetting('VK_RIGHT',inttostr(ord('P'))));
   fire:=StrToInt(GetSetting('VK_FIRE',inttostr(ord( ' '))));
  end;
    SetDisplay(glDisplayMode,glRealBorder,glEmulatedFlash,glFrameSkip);
    SetDisplaySize(glDisplayWidth, glDisplayHeight);

{$IFDEF USE_SOUND}
        if assigned(AY8912) then AY8912.Destroy;
        AY8912:=TAY8912.Create(1773000, WAVE_FREQUENCY, 8);
{$ENDIF}
  if assigned(Tape) then tape.destroy;
  tape:=ttapeinfo.create;

  if assigned(Betadisk) then betadisk.destroy;
  betadisk:=tbetadisk.create;

    glInterruptDelay := StrToInt(GetSetting('InterruptDelay', '20'));

    // Initialize everything
    initParity;
    Z80Reset;

    glResolution:=StrToInt(GetSetting('Resolution','0'));
    glAutoreboot:=StrToInt(GetSetting('Autoreboot','0'))>0;
    glEmulatedMF128:=StrToInt(GetSetting('Multiface','1'))>0;
    glBetaDisk:=StrToInt(GetSetting('Betadisk','0'))>0;
    SetEmulatedModel(StrToInt(GetSetting('EmulatedModel', '0')));

    resetKeyboard;

    timeBeginPeriod(1);

    glInterruptTimer := timeGetTime() + 20;
    {$IFDEF USE_SOUND}
        gbSoundEnabled := InitializeWaveOut();

     gbSpeakerenabled:= GetSetting('SpeakerEnabled','1')='1';
     gbAY8912enabled:= GetSetting('AY8912Enabled','1')='1';

        If gbSoundEnabled Then
        begin
            For ii := 1 To 12 do
              waveOutWrite(glphWaveOut, @gtWavHdr[ii], sizeof(WAVEHDR));

            glBufNum := 12;
        End;
    {$ENDIF}
End;

procedure resetKeyboard;
begin
    keyB_SPC := $FF;
    keyH_ENT := $FF;
    keyY_P := $FF;
    key6_0 := $FF;
    key1_5 := $FF;
    keyQ_T := $FF;
    keyA_G := $FF;
    keyCAPS_V := $FF;
End;

procedure SetKempston(Up,Down,Left,Right,Fire: TShortCut);
begin
 gljoystick.up:=up;
 gljoystick.down:=down;
 gljoystick.left:=left;
 gljoystick.right:=right;
 gljoystick.fire:=fire;
 SaveSetting('VK_UP', IntToStr(Up));
 SaveSetting('VK_DOWN', IntToStr(Down));
 SaveSetting('VK_LEFT', IntToStr(Left));
 SaveSetting('VK_RIGHT', IntToStr(Right));
 SaveSetting('VK_FIRE', IntToStr(Fire));
end;

procedure SetDisplay(Mode:Byte;Border,Flash:Boolean;Skip:Byte);
var oMode:Byte;
    oborder:Byte;
    fMode: TMode;
begin
// oMode:=Byte(frmMainWnd.Display.mode);
// if Mode=oMode then exit; // same as before
 glDisplayMode:=Mode;
 glRealBorder:=Border;
 SaveSetting('DisplayMode', IntToStr(glDisplayMode));
 SaveSetting('DisplayBorder', IntToStr(Byte(glRealborder)));
 SaveSetting('DisplayFrameSkip',IntToStr(Byte(glFrameSkip)));
 SaveSetting('EmulatedFlash', IntToStr(byte(glEmulatedFlash)));
 with frmMainWnd do
 begin
  if assigned(display) then
   begin
    oborder:=Display.Border;
    Display.Destroy;
   end;

  if Mode=0 then fmode:=GDI else
  if Mode=1 then fmode:=DDraw else fMode:=OVL;
  Display:=TOutput.create(Handle,glRealBorder,fMode);
  Display.Border:=oborder;
  Display.Skip:=Skip;
  Display.Flash:=GLEmulatedFlash;
  formresize(nil);
  screenPaint;
 end;
end;

procedure SetDisplaySize(lWidth : integer; lHeight : integer);
var mult,Y:byte;
begin
// with frmMainWnd do y:=Toolbar.Height+Statusbar.Height;
  glDisplayWidth := lWidth;
  glDisplayHeight := lHeight;
  SaveSetting('DisplayWidth', IntToStr(glDisplayWidth));
  SaveSetting('DisplayHeight', IntToStr(glDisplayHeight));

 With frmMainWnd do
 if (WindowState=wsNormal) then
 Begin
  Case lWidth of
    256:  // Standard
      begin
        ClientWidth := 256;
        ClientHeight := 192+Y;
        mult:=1;
      end;
    512:
      begin
        ClientWidth := 512;
        ClientHeight := 384+Y;
        mult:=2;
      end;
    768:
      begin
        ClientWidth := 768;
        ClientHeight := 576+Y;
        mult:=3;
      end;
   End;
   if glRealBorder then
    begin
     clientwidth:=clientwidth+(mult*32) shl 1;
     clientheight:=clientheight+(mult*32) shl 1;
    end;
  End;
End;


procedure HookTape(Page:Byte);
begin
 gRAMPage[Page,1386]:=237;
 gRamPage[Page,1387]:=255;
end;

function GetRom(Count:integer):String;
begin
 if assigned(frmsetup) and (frmsetup.lvroms.items.count>count) then
   result:=frmsetup.lvroms.items[count].subitems[0]
  else
   result:=stdroms[count].n;

  if pos('\',result)=0 then // local !!
   result:=GetAppPath+result
end;

procedure SetEmulatedModel(lModel : integer);
var sModel : String;
begin
    glEmulatedModel := lModel;
    glEmulatedTRDos := lModel in [5..6]; // Pentagon&Scorpion
    Case lModel of
      0:
        begin
          // A 48K Spectrum has 69888 tstates per interrupt (3.50000 MHz)
          sModel := 'ZX Spectrum 48K';
          glTstatesPerInterrupt := 69888;
          glTstatesPerLine:=224;

          {$IFDEF USE_SOUND}
          glWaveAddTStates := 158;
          {$ENDIF}
          glUseScreen := 5;

          glPageAt[0] := nRomStart;
          glPageAt[1] := 5;
          glPageAt[2] := 1;
          glPageAt[3] := 2;
          glPageAt[4] := nRomStart;
          // load the ROM image into memory
          if LoadROM(GetRom(0) {'spectrum.rom'}, nRomStart) = false then
            Application.Terminate;
          HookTape(nRomStart);
          outb:=zx48outb;
          inb:=zx48inb;
        end;
      1:
        begin
          // A 128K Spectrum has 70908 tstates per interrupt (3.54690 MHz)

          sModel := 'ZX Spectrum 128';
          glTstatesPerInterrupt := 70908;
          glTstatesPerLine:=228;
          {$IFDEF USE_SOUND}
          glWaveAddTStates := 160;
          {$ENDIF}

          glUseScreen := 5;

          glPageAt[0] := nRomStart;
          glPageAt[1] := 5;
          glPageAt[2] := 2;
          glPageAt[3] := 0;
          glPageAt[4] := nRomStart;

           if LoadROM(Getrom(2){'zx128_1.rom'}, nRomStart+1) = false then
            Application.Terminate;
           if LoadROM(Getrom(1){'zx128_0.rom'}, nRomStart) = false then
             Application.Terminate;
           HookTape(nRomStart+1);

          outb:=zx128outb;
          inb:=zx128inb;
        end;
      2:
        begin
          // A Spectrum +2 has 70908 tstates per interrupt (3.54690 MHz)
          sModel := 'ZX Spectrum +2';
          glTstatesPerInterrupt := 70908;
          glTstatesPerLine:=228;
          {$IFDEF USE_SOUND}
          glWaveAddTStates := 160;
          {$ENDIF}
          glUseScreen := 5;

          glPageAt[0] := nRomStart;
          glPageAt[1] := 5;
          glPageAt[2] := 2;
          glPageAt[3] := 0;
          glPageAt[4] := nRomStart;
          if LoadROM(GetRom(4){'plus2_1.rom'}, nRomStart+1) = false then
            Application.Terminate;
          if LoadROM(GetRom(3){'plus2_0.rom'}, nRomStart) = false then
            Application.Terminate;
          HookTape(nRomStart+1);
          outb:=Plus2outb;
          inb:=Plus2inb;
        end;
      3:  // +2A
        begin
          sModel := 'ZX Spectrum +2A';
          glTstatesPerInterrupt := 70908;
          glTstatesPerLine:=228;
          {$IFDEF USE_SOUND}
          glWaveAddTStates := 160;
          {$ENDIF}

          glUseScreen := 5;

          glPageAt[0] := nRomStart;
          glPageAt[1] := 5;
          glPageAt[2] := 1;
          glPageAt[3] := 2;
          glPageAt[4] := nRomStart;

          if LoadROM(GetRom(8){'plus2a_3.rom'}, nRomStart+3) = false then
            Application.Terminate;
          if LoadROM(GetRom(7){'plus2a_2.rom'}, nRomStart+2) = false then
            Application.Terminate;
          if LoadROM(GetRom(6){'plus2a_1.rom'}, nRomStart+1) = false then
            Application.Terminate;
          if LoadROM(GetRom(5){'plus2a_0.rom'}, nRomStart) = false then
            Application.Terminate;
          HookTape(nRomStart+1);
          outb:=Plus3outb;
          inb:=Plus3inb;

        end;
      4:  // +3
        begin
          sModel := 'ZX Spectrum +3';
          glTstatesPerInterrupt := 70908;
          glTstatesPerLine:=228;
          {$IFDEF USE_SOUND}
          glWaveAddTStates := 160;
          {$ENDIF}

          glUseScreen := 5;

          glPageAt[0] := nRomStart;
          glPageAt[1] := 5;
          glPageAt[2] := 1;
          glPageAt[3] := 2;
          glPageAt[4] := nRomStart;

          if LoadROM(GetRom(12){'plus3_3.rom'}, nRomStart+3) = false then
            Application.Terminate;
          if LoadROM(GetRom(11){'plus3_2.rom'}, nRomStart+2) = false then
            Application.Terminate;
          if LoadROM(GetRom(10){'plus3_1.rom'}, nRomStart+1) = false then
            Application.Terminate;
          if LoadROM(GetRom(9){'plus3_0.rom'}, nRomStart) = false then
            Application.Terminate;
          HookTape(nRomStart+1);
          outb:=Plus3outb;
          inb:=Plus3inb;
        end;
      5:  // Pentagon
        begin
          sModel := 'Pentagon 128k';
          glTstatesPerInterrupt := 70908;
          glTstatesPerLine:=228;
          {$IFDEF USE_SOUND}
          glWaveAddTStates := 160;
          {$ENDIF}
          glUseScreen := 5;

          glPageAt[0] := nRomStart;
          glPageAt[1] := 5;
          glPageAt[2] := 2;
          glPageAt[3] := 0 ;
          glPageAt[4] := nRomStart;

          if LoadROM(GetRom(19) {'trdos.rom'}, nRomStart+3) = false then
            Application.Terminate;
          if LoadROM(GetRom(14) {'zx128_1.rom'}, nRomStart+1) = false then
            Application.Terminate;
          if LoadROM(GetRom(13) {'128tr.rom'}, nRomStart) = false then
            Application.Terminate;
          HookTape(nRomStart+1);
          outb:=Pentagonoutb;
          inb:=Pentagoninb;
        end;
      6:  // Scorpion
        begin
          sModel := 'Scorpion 256k';
          glTstatesPerInterrupt := 70908;
          glTstatesPerLine:=224;
          {$IFDEF USE_SOUND}
          glWaveAddTStates := 160;
          {$ENDIF}
          glUseScreen := 5;

          glPageAt[0] := nRomStart;
          glPageAt[1] := 5;
          glPageAt[2] := 2;
          glPageAt[3] := 0;
          glPageAt[4] := nRomStart;

          if LoadROM(GetRom(18){'scorp3.rom'}, nRomStart+3) = false then
            Application.Terminate;
          if LoadROM(GetRom(17){'scorp2.rom'}, nRomStart+2) = false then
            Application.Terminate;
          if LoadROM(GetRom(16){'scorp1.rom'}, nRomStart+1) = false then
            Application.Terminate;
          if LoadROM(GetRom(15){'scorp0.rom'}, nRomStart) = false then
            Application.Terminate;
          //HookTape(nRomStart+1);
          outb:=Scorpionoutb;
          inb:=Scorpioninb;
        end;
    end;
   if glEmulatedMF128 then
   begin
    if lmodel in [3,4] then
     LoadROM(GetRom(21){'mfplus3.rom'}, nMF128start)
    else
     LoadROM(GetRom(20){'mf128.rom'}, nMF128start);
   end;
   if (glBetaDisk) and (not (lmodel in [5,6])) then
   begin
    if LoadROM(GetRom(19){'trdos.rom'}, nRomStart+3) = false then
     Application.Terminate;
    glEmulatedTRDos:=true;
   end;
   SetStatus(sModel);
End;


procedure SetStatus(sMsg : String);
begin
 frmMainWnd.statusbar.panels[0].text:=sMsg;
 {$IFDEF USE_DEBUGGING}
   ReportDebug(0, 'Setmodel '+sMsg, pc.w, false);
 {$ENDIF}

End;

end.






