unit ModBetadisk;

interface uses windows;

type
    SCL_Header=record
             filename:array[0..7] of char;
             filetype:byte;
             Size_lo,Size_Hi:Byte;
             Len_lo,Len_Hi:Byte;
             Size:Byte;
            end;

    tbyte=Array[word]of byte;
    pbyte=^tbyte;
    tWord=Array[word]of word;
    pword=^tword;

    TDiskIO=Class
             private
              fdisk: Pointer;
              fsize: Dword;
              fposition:DWord;
              ffilename:String;
              fchanged:Boolean;

              procedure SetFileName(Value:String);
              procedure ReadSCL(Filename:String);
              procedure ReadHob(Filename:String);
              procedure AddFile(Hdr:SCL_HEADER;Var Data);
             public
              constructor Create;
              destructor Destroy;
              procedure EmptyDisk;
              procedure Eject;
              procedure Read(Var Buffer;Size:DWord;Out NumRead:DWord);
              procedure Write(Var Buffer;Size:DWord;Out NumWritten:DWord);
              procedure Seek(Position: DWord);

              property FilePos:DWord read fPosition;
              property FileSize: Dword read fSize;
              property Filename: String read FFilename write SetFilename;
            end;

    TBetadisk=class
               private
                  RegStatus,
                  RegCom:Byte;
                  RegTrack:Shortint;
                  RegSect,
                  RegData,
                  System:Byte;
                  StepDirect:Shortint;
                  TrackReal:array[0..3] of byte;
                  reserved:array[0..2] of byte;
                  DiskBuf: array[0..$FF] of byte;
                  Current_Disk: Byte;
                  DskIndexCounter,
                  DskCountDatLost: Byte;
                  CntData,DskDataSize: Integer;
                  CntReady: Word;
                  function getfa(S:String):integer;
                  function DiskPosition(Sect:Word):Longint;
                  procedure SetStateStep;
                  procedure OpCom;
                  procedure StepRear;
                  procedure OpReadSector;
                  procedure OpWriteSector;
               public
                Disks:array[0..3] of TDiskIO;
                procedure DiskVG(OperIO:Byte;Var DataIO:Byte);
                constructor create;
                destructor destroy;
               end;

var betadisk: TBetadisk;

implementation uses dialogs,sysutils,agopendialog;

{$i boot.inc}

const
      FormatDatLen             = 1;
      Ready          : Boolean = True;
      ReadIndexTrack : Boolean = False;
      DskStatus      : Byte    = 00;

      DskIndex       : array[0..4] of byte = (0,1,1,0,0);

constructor TDiskIO.create;
begin
 inherited create;
 fdisk:=nil;fsize:=0;
 fchanged:=false;
 fposition:=0;ffilename:='';
end;

destructor TDiskIO.destroy;
begin
 Eject;
end;

procedure TDiskIO.Eject;
var str:File;
    dlg:TagOpenDialog;
begin
 if (fchanged) and
    (MessageDlg('Do you want to save the disk?',
    mtConfirmation, [mbYes, mbNo], 0) = idYes) then
     begin
      dlg:=TagOpenDialog.Create(nil);
      dlg.filter:='TRD files (*.trd)|*.trd';
      if dlg.execute then ffilename:=dlg.filename;
      dlg.Destroy;
      {$i-}
      assignfile(str,filename);
      rewrite(str,1);
      blockwrite(str,fdisk^,fsize);
      closefile(str);
      {$i+}
     end;
 if fdisk<>nil then freemem(fdisk,fsize);
 fdisk:=nil;
end;

procedure TDiskIO.EmptyDisk;
begin
 fchanged:=false;
 if fdisk<>nil then freemem(fdisk,fsize);

 fsize:=655360;//256*4096;
 getmem(fdisk,fsize);
 zeromemory(fdisk,fsize);
 pbyte(fdisk)^[$08E2]:= $1;    // first free track
 pbyte(fdisk)^[$08E3]:= $16;   // 80T,DS
 pword(fdisk)^[$08E5]:= 2544;  // free sec
 pbyte(fdisk)^[$08E7]:= $10;   // trdos flag
end;

procedure TDiskIO.ReadSCL(Filename:String);
var str:File;
    temp:PChar;
    sz:Dword;
    id:Array[0..7] of char;
    data:DWord;
    files,i:integer;
    hdr:SCL_Header;
    autoboot:Boolean;
begin
 temp:=nil;
 {$i-}
   assignfile(str,Filename);filemode:=0;
   reset(str,1);
   if ioresult=0 then
   begin
    sz:=system.filesize(str); // SCL_Size
    getmem(temp,sz);
    blockread(str,temp^,sz); // Read whole SCL-File
   end;
   closefile(str);
 {$i+}
 emptydisk;
 if temp=nil then exit;
 move(temp^,ID,8);// SINCLAIR
 if ID<>'SINCLAIR' then exit;
 files:=byte(temp[8]);
 data:= 9 + 14*files;
 autoboot:=not true;//false;
 for i:= 0 to files-1 do
  begin
   move(temp[9+14*i],hdr,14);
   if hdr.filename='boot    B' then autoboot:=false;
   addfile(hdr, temp[data]);
   data:=data + (9 + 14*i + 13)*$100;
   end;
  freemem(temp,sz);
  if autoboot then
   begin
    getmem(temp,sizeof(boot));
    move(boot,temp^,sizeof(boot));
    temp[13]:= temp[14];
    move(temp^,hdr,14);
    addfile(hdr, temp[17]);
    freemem(temp,sizeof(boot));
   end;
  {assignfile(str,'demo.trd');rewrite(str,1);
  blockwrite(str,fdisk^,fsize);
  closefile(str);}
end;

procedure TDiskIO.AddFile(Hdr:SCL_HEADER;Var Data);
var pos:integer;
begin
   if pword(fdisk)^[$08E5]< hdr.Size then exit; // disk full
   pword(fdisk)^[$08E5]:= pbyte(fdisk)^[$08E5]- hdr.Size;
   pos:= pbyte(fdisk)^[$08E1] + 16*pbyte(fdisk)^[$08E2];
   move(data,pbyte(fdisk)^[pos*$100],hdr.Size*$100);
   move(hdr,pbyte(fdisk)^[pbyte(fdisk)^[$08E4]*$10],14);

   pbyte(fdisk)^[pbyte(fdisk)^[$08E4]*$10+14]:= pbyte(fdisk)^[$08E1];
   pbyte(fdisk)^[pbyte(fdisk)^[$08E4]*$10+15]:= pbyte(fdisk)^[$08E2];
   inc(pos,hdr.Size);
   pbyte(fdisk)^[$08E1]:= pos and $0F;
   pbyte(fdisk)^[$08E2]:= pos shr 4;
   pbyte(fdisk)^[$08E4]:= pbyte(fdisk)^[$08E4]+1;
end;

procedure TDiskio.ReadHOB(Filename:String);
var str:file;
    temp:Pchar;
    hdr:scl_header;
    sz:dword;
begin
 temp:=nil;
 {$i-}
   assignfile(str,Filename);filemode:=0;
   reset(str,1);
   if ioresult=0 then
   begin
    sz:=system.filesize(str); // SCL_Size
    getmem(temp,sz);
    blockread(str,temp^,sz); // Read whole SCL-File
   end;
   closefile(str);
 {$i+}
 temp[13]:= temp[14];
 move(temp^,hdr,14);
 addfile(hdr, temp[17]);
 freemem(temp,sz);
end;

procedure TDiskIO.SetFileName(Value:String);
var str:file;
begin
 if ffilename=Value then exit;
 Eject;
 ffilename:=Value;
 fchanged:=false;
 fposition:=0;

 if (ffilename='') or (not fileexists(ffilename)) then EmptyDisk else
 begin
  if lowercase(extractfileext(value))='.scl' then
   begin
    ReadSCL(Value);
    exit;
   end else
   if pos('$',lowercase(extractfileext(value)))>0 then
    begin
     EmptyDisk;
     readHOB(Value);
     exit;
    end;
  {$i-}
  assignFile(str,Filename);
  filemode:=0; reset(str,1);
  if ioresult=0 then
    begin
     fsize:=system.filesize(str); // Get Disksize
     getmem(fdisk,fsize);
     blockread(str,fdisk^,fsize); // Read whole Disk
    end else EmptyDisk;
   closefile(str);
  {$i+}
 end;
end;

procedure TDiskIO.Read(Var Buffer;Size:DWord;Out NumRead:DWord);
begin
 if fposition+size>fsize then numread:=fsize-fposition else numread:=size;
 move(pointer(dword(fdisk)+fposition)^,buffer,numread);
end;

procedure TDiskIO.Write(Var Buffer;Size:DWord;Out NumWritten:DWord);
begin
 if fposition+size>fsize then numwritten:=fsize-fposition else numwritten:=size;
 move(buffer,pointer(dword(fdisk)+fposition)^,numwritten);
 fchanged:=true;
end;

procedure TDiskIO.Seek(Position: DWord);
begin
 fposition:=Position;
 if fposition>fsize then fposition:=fsize;
end;

function TBetadisk.getfa(S:String):integer;
begin
 if fileexists(s) then result:=$20 else result:=-1;
end;

function TBetadisk.DiskPosition(Sect:Word):Longint;
begin
  result:=((RegTrack shl 5) or (System and $10 xor $10) or Sect) shl $08;
end;

procedure TBetadisk.SetStateStep;
begin
        if RegTrack<0 then
              begin
                TrackReal[System and $03]:=0;
                RegTrack:=0;
              end;
        if RegTrack>$4F then
              begin
                TrackReal[System and $03]:=$4F;
                RegTrack:=$4F;
              end;

        RegStatus:=DskStatus and $40;
        if RegCom and $08>0 then RegStatus:=RegStatus or $20;
        if RegTrack=0 then RegStatus:=RegStatus or $04;
end;

procedure TBetadisk.OpCom;
begin
 DskCountDatLost:=$00;
 CntData:=$0000;
 RegStatus:=$03;
 if not Ready then
  begin
   RegStatus:=$01;
   CntReady:=$10;
  end;
end;

procedure TBetadisk.StepRear;
begin
 if RegTrack>0 then
  begin
   dec(RegTrack);
   TrackReal[System and $03]:=RegTrack;
  end;
  StepDirect:=-1;
  SetStateStep;
end;

procedure TBetadisk.OpReadSector;
var numread:DWord;
    Position:longint;
begin
 if (RegSect>$10) or (RegTrack<>TrackReal[System and $03]) or
    (DskStatus and $04=0) then
  begin
   RegStatus:=$10; {OpNotFound}
   exit;
  end;

 Position:=DiskPosition(RegSect-1);

 disks[Current_Disk].Seek(Position);
 disks[Current_Disk].Read(diskbuf,sizeof(diskbuf),numread);

 if Position=2048 then diskbuf[227]:=16; {BUG !!! at pos $8e3 must be $10 ??}

 if numread<>sizeof(diskbuf) then RegStatus:=$08 {OpContSum}
  else
   begin
    DskDataSize:=$FF;
    RegData:=DiskBuf[0];
    OpCom;
   end;
end;

procedure TBetadisk.OpWriteSector;
begin
 if DskStatus and $40>0 then
  begin
   RegStatus:=$40; {OpWrProt}
   exit;
  end;
 if (RegSect>$10) or (RegTrack<>TrackReal[System and $03]) or
    (DskStatus and $04=0) then RegStatus:=$10 {OpNotFound}
   else
   begin
    DskDataSize:=$0100;
    OpCom;
   end;
end;

procedure TBetadisk.DiskVG(OperIO:Byte;Var DataIO:Byte);

var numread,numwritten:DWord;
    i,attr:integer;
    IntrqDrq:Word;
    Intrq,Drq:Byte;
begin
        if RegStatus and $01>0 then
         begin
          dec(DskCountDatLost);
          if dskcountdatlost=0 then RegStatus:=$04;
         end;
        if CntReady>0 then
         begin dec(CntReady);if (CntReady=0) then RegStatus:=$03; end;

        case OperIO of
         $0: begin
                DataIO:=RegStatus;
                if RegCom and $80>0 then exit;
                dec(DskIndexCounter);
                if DskIndexCounter and $E>0 then DataIO:=DataIO or $02;
              end;
         $1: begin
                if RegStatus and $01=0 then
                begin
                        if (DataIO and $F0)=$D0 then exit;
                        RegCom:=DataIO;
                        case RegCom shr 4 of
                         $0: begin
                                TrackReal[System and $03]:=0;
                                RegTrack:=0;
                                StepDirect:=-1;
                                SetStateStep;
                              end;
                         $1: begin
                                TrackReal[System and $03]:=RegData;
                                RegTrack:=RegData;
                                if RegData-RegTrack<0 then StepDirect:=-1;
                                if RegData-RegTrack>0 then StepDirect:=1;
                                SetStateStep;
                             end;
                          $2,
                          $3,
                          $4,
                          $5:begin
                              if ( (RegCom shr 4) in [2,3]) and (StepDirect=-1) then
                               begin StepRear;exit;end;
                                if RegTrack>0 then
                                 begin
                                  inc(RegTrack);
                                  TrackReal[System and $03]:=RegTrack;
                                 end;
                                StepDirect:=1;
                                SetStateStep;
                              end;
                         $6,
                         $7: StepRear;
                         $8,
                         $9: OpReadSector;
                         $A,
                         $B: OpWriteSector;
                         $C: begin
                                DskDataSize:=$0005;
                                RegData:=1;
                                if ReadIndexTrack then RegData:=RegTrack;
                                OpCom;
                             end;
                         $E:;
                         $F: begin
                                if DskStatus and $40>0 then
                                  RegStatus:=$40 {OpWrProt}
                                else
                                begin
                                 DskDataSize:=FormatDatLen;
                                 OpCom;
                                end;
                             end;
                        end;
                end;
               end;
         $2: DataIO:=RegTrack;
         $3: if RegStatus and $01=0 then RegTrack:=DataIO;
         $4: DataIO:=RegSect;
         $5: if RegStatus and $01=0 then RegSect:=DataIO;
         $6: begin
                if RegStatus and $01>0 then
                 begin
                        case RegCom shr 4 of
                         $8,
                         $9: begin
                                DataIO:=RegData;
                                DskCountDatLost:=$10;
                                if CntData<DskDataSize then
                                 begin
                                  inc(cntData);
                                  RegData:=DiskBuf[CntData];
                                 end
                                else begin
                                        if RegCom and $10>0 then
                                         begin
                                                inc(RegSect);
                                                OpReadSector;
                                         end else RegStatus:=$00; {OpOk}
                                end
                              end;
                          $C: begin
                                DataIO:=RegData;
                                DskCountDatLost:=$10;
                                if CntData<DskDataSize then
                                 begin
                                  RegData:=DskIndex[CntData];
                                  inc(CntData);
                                 end
                                else RegStatus:=$00; {OpOk}
                               end;
                        end
                end else DataIO:=RegData;
                end;
         $7: begin
                if (RegStatus and $01>0) then
                 begin
                        case RegCom shr 4 of
                         $A,
                         $B: begin
                                DiskBuf[CntData]:=DataIO;
                                RegData:=DataIO;inc(CntData);
                                DskCountDatLost:=$10;
                                if CntData>=DskDataSize then
                                 begin
                                        disks[Current_Disk].Seek(DiskPosition(RegSect-1));
                                        disks[Current_Disk].Write(Diskbuf,sizeof(Diskbuf),numwritten);
                                        if numwritten<>sizeof(DiskBuf) then
                                         begin
                                          RegStatus:=$08; {OpContSum}
                                          exit;
                                         end;
                                        if (RegCom and $10>0) then
                                         begin
                                                inc(RegSect);
                                                OpWriteSector;
                                        end else RegStatus:=$00; {OpOk}
                                end
                              end;
                          $F: begin
                                RegData:=DataIO;
                                DskCountDatLost:=$10;
                                inc(cntData);
                                if CntData>=DskDataSize then
                                begin
                                        fillchar(DiskBuf,sizeof (DiskBuf),0);
                                        if DskStatus and $04>0 then
                                        begin
                                                disks[Current_Disk].Seek(DiskPosition(0));
                                                for i:=0 to $10-1  do disks[Current_Disk].Write(DiskBuf,sizeof(Diskbuf),NumWritten);
                                        end else
                                        begin
                                         // disks[Current_Disk].EmptyDisk;
                                         // Problem!!!!!!!!11
                                         for i:=0 to $A00-1 do disks[Current_Disk].Write(DiskBuf,sizeof(Diskbuf),NumWritten);
                                         DskStatus:=DskStatus or $4;
                                        end;
                                        RegStatus:=$00; {OpOk}
                                end;
                           end;
                       end
                end else
                        RegData:=DataIO;
          end;
        $8: begin
                {if RegStatus and 32>0 then DRQ:=$40 else DRQ:=0;
                if Regstatus and 64>0 then Intrq:=$80 else Intrq:=0;
                Dataio:=DRQ or intrq;}
                IntrqDrq:=$BF;
                if RegStatus and $01>0 then IntrqDrq:=$3F;
                if RegStatus and $02>0 then IntrqDrq:=$7F;
                DataIO:=IntrqDrq;
             end;
         $9: begin
                if (RegStatus and $01>0) and (RegCom and $80>0) then
                 begin
                  RegStatus:=$08; {OpContSum}
                  exit;
                 end;
                System:=DataIO;
                if (System xor DskStatus) and $03>0 then
                 begin
                        DiskVG($B,DataIO);
                        DiskVG($A,DataIO);
                 end;
               end;
          $A: begin
                if DskStatus and $80>0 then exit;
                DskStatus:=System and $03;
                Current_Disk:=DskStatus;

                Attr:=GetFA(disks[Current_Disk].FileName);
                if Attr<>-1 then
                begin
                        Attr:=Attr and $01;
                        DskStatus:=DskStatus or (Attr shl 6);
                        disks[Current_Disk].seek(0); // Reset
                        if ioresult=0 then DskStatus:=DskStatus or $04;
                end;
                DskStatus:=DskStatus or $80;
               end;
         $B: begin
                if DskStatus and $80=0 then exit;
                // if DskStatus and $04>0 then disks[Current_Disk].eject;//close(Str); !!!!
                DskStatus:=$00;
              end;
         $C: begin
                RegStatus:=$24;
                RegCom:=$00;
                RegTrack:=$00;
                RegSect:=$01;
                RegData:=$00;
                System:=$3C;
                StepDirect:=shortint($FF);
                if (DskStatus and $80>0) and (DskStatus and $03>0) then
                 begin
                        DiskVG($B,DataIO);
                        DiskVG($A,DataIO);
                 end;
            end;
        end;
end;


constructor TBetadisk.create;
var i:byte;
begin
 Current_Disk:=0;
 for i:=low(disks) to high(disks) do
  Disks[i]:=TDiskIO.Create;
end;

destructor TBetaDisk.destroy;
var i:byte;
begin
 for i:=low(disks) to high(disks) do Disks[i].Destroy;
end;

end.
