{' /*******************************************************************************
'   modLoader.pas within DelphiSpec.dpr
'
'   Routines for loading the ROM, snapshots, Tapes and Discs
'
'   This program is free software; you can redistribute it and/or
'   modify it under the terms of the GNU General Public License
'   as published by the Free Software Foundation; either version 2
'   of the License, or (at your option) any later version.
'   This program is distributed in the hope that it will be useful,
'   but WITHOUT ANY WARRANTY; without even the implied warranty of
'   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
'   GNU General Public License for more details.
'
'   You should have received a copy of the GNU General Public License
'   along with this program; if not, write to the Free Software
'   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
'
' *******************************************************************************/}
// $Todo: TAP_save, TZX,VOC support & several other formats
unit modLoader;

interface uses modmain;

function LoadROM(sROMFile : String = 'spectrum.rom'; lROMPage : integer = nromstart): boolean;

procedure LoadZ80Snap(sFilename : String);
procedure LoadSNASnap(sFilename : String);
procedure SaveROM(sFilename : string);
procedure SaveSNASnap(sFilename : String);
procedure SaveZ80Snap(sFilename : String);

implementation
uses SysUtils,Dialogs, Windows,
               ModZ80, ModAY8912,ModDisplay,UMain;


Function GetFilePart(sFilename : String) : String;
begin
    GetFilePart := ExtractFileName(sFileName);
End;

//Return value True=Success, False=Failure
function LoadROM(sROMFile : String = 'spectrum.rom'; lROMPage : integer = nromstart): boolean;
  var
    hFile : File;
    ret: integer;
    sROM : array[0..65536] of Byte;
    lCounter : integer;
    sr: TSearchRec;

  begin
    LoadROM := True;
    If FindFirst(sROMFile, 0, sr) <> 0 Then
    begin
      MessageDLG(
        'The ROM image file "' + sROMFile + '" could not be found.',
        mtinformation,[mbOk],0);

      LoadROM := false;
      Exit;
    End;

    AssignFile(hFile, sRomFile);
    filemode:=0;
    reset(hFile, 1);

    // Read the ROM image into sROM
    Blockread(hFile, sROM, 16384, ret);
     
    If (ret<>8192) and (ret<>16384) then
    begin
     MessageDLG(
        'An error ocurred whilst reading the ROM image file: "' + sROMFile + '"',
        mtError,[mbok],0);

      CloseFile(hFile);

      LoadRom := false;
      Exit;
    End
    else
    begin
{$IFDEF USE_DEBUGGING}
      ReportDebug(0,'LoadROM Start, Page ' + IntToStr(lROMPage), 0, True);
{$ENDIF}
      // Copy the ROM into the appropriate memory page
      For lCounter := 1 To 16384 do
      begin
        gRAMPage[lROMPage, lCounter - 1] := sROM[lCounter-1];
      end;
      CloseFile(hFile);
{$IFDEF USE_DEBUGGING}
      ReportDebug(0,'LoadROM End, Page ' + IntToStr(lROMPage), 0, True);
{$ENDIF}
    end;
end;

procedure LoadSNA128Snap(var hFile : File);
  var
    sData : array[0..65535] of byte;
    sTemp : array[0..1] of byte;
    lOut7FFD : integer;
    lBank : integer;
    lCounter : integer;
begin
    // Read first three banks
    BlockRead(hFile, sData, 49152);

    // PC
    BlockRead(hFile, sTemp, 2);
    PC.W := 256*sTemp[1]+sTemp[0];

    // Last out to 0x7FFD
    BlockRead(hFile, sTemp, 1);
    lOut7FFD := sTemp[0];

    // Is TR-DOS paged? (ignored by vbSpec)
    BlockRead(hFile, sTemp, 1);

    //Setup first three banks
    For lCounter := 0 To 16383 do
    begin
        gRAMPage[5, lCounter] := sData[lCounter];
        gRAMPage[2, lCounter] := sData[lCounter + 16384];
        gRAMPage[lOut7FFD And 7, lCounter] := sData[lCounter + 32768];
    end;

    lBank := 0;
    While lBank < 8 do
    begin
        BlockRead(hFile, sData, 16384);
        If (lBank = 5) Or (lBank = 2) Or (lBank = (lOut7FFD And 7)) Then
            lBank := lBank + 1;

        If lBank < 8 Then
        begin
            For lCounter := 0 To 16383 do
                gRAMPage[lBank, lCounter] := sData[lCounter];

            lBank := lBank + 1;
        End;
    end;

    outb($7FFD, lOut7FFD);
End;

procedure LoadSNASnap(sFilename : String);
    var hFile : File; sData : array[0..65535] of byte; iCounter : integer; temp:integer;

begin
    AssignFile(hFile, sFileName);
    filemode:=0;
    reset(hFile, 1);
    BlockRead(hFile, sData, 1);
    IR.H := sData[0];

    BlockRead(hFile, sData, 2);
    HL2.w := 256*sData[1]+sData[0];

    BlockRead(hFile, sData, 2);
    DE2.W := 256*sData[1]+sData[0];

    BlockRead(hFile, sData, 2);
    BC2.W := 256*sData[1]+sData[0];

    BlockRead(hFile, sData, 2);
    AF2.W := 256*sData[1]+sData[0];

    BlockRead(hFile, sData, 2);
    HL.W := 256*sData[1]+sData[0];

    BlockRead(hFile, sData, 2);
    DE.W := 256*sData[1]+sData[0];

    BlockRead(hFile, sData, 2);
    BC.W:=256*sData[1]+sData[0];

    BlockRead(hFile, sData, 2);
    IY.W := 256*sData[1]+sData[0];

    BlockRead(hFile, sData, 2);
    ix.w := 256*sData[1]+sData[0];

    BlockRead(hFile, sData, 1);
    If (sData[0] And 4) = 4 Then
    begin
      IFF1 := 1;
      IFF2 := 1;
    end
    Else
    begin
      IFF1 := 0;
      IFF2 := 0;
    End;

    BlockRead(hFile, sData, 1);
    IR.l := sData[0];
    RTemp := IR.L;

    BlockRead(hFile, sData, 2);
    AF.W:=256*sData[1]+sData[0];

    BlockRead(hFile, sData, 2);
    SP.W := 256*sData[1]+sData[0];

    BlockRead(hFile, sData, 1);
    IM := sData[0];

    BlockRead(hFile, sData, 1);
    // Border color
    glNewBorder := sData[0] And $7;

    Temp := FileSize(hFile);
    If Temp > 49180 Then
    begin
        LoadSNA128Snap(hFile);
        CloseFile(hFile);

        // Set the initial border color
        display.border := glNewBorder;

        SetEmulatedModel(1);

        screenPaint;
        resetKeyboard;
    end
    Else
    begin
        // Load a 48K Snapshot file
        BlockRead(hFile, sData, 49152);
        CloseFile(hFile);

{$IFDEF USE_DEBUGGING}
        ReportDebug(0,'LoadSNA_0', 0, True);
{$ENDIF}
        For iCounter := 0 To 16383 do
        begin
            gRAMPage[5, iCounter] := sData[iCounter];
            gRAMPage[1, iCounter] := sData[iCounter + 16384];
            gRAMPage[2, iCounter] := sData[iCounter + 32768];
        end;

{$IFDEF USE_DEBUGGING}
        ReportDebug(0,'LoadSNA_1', 0, True);
{$ENDIF}
        // Set the initial border color
        display.border:= glNewBorder;

        SetEmulatedModel(0);
{$IFDEF USE_DEBUGGING}
        ReportDebug(0,'LoadSNA_2', 0, True);
{$ENDIF}

        screenPaint;
        resetKeyboard;
        pop(pc.w);
    End;
End;



procedure ReadZ80V2orV3Snap(var hFile : File);
  var
  lHeaderLen : integer; sData : array[0..65535] of byte;
  lCounter : integer; bHardwareSupported : Boolean;
  lMemPage : integer; lBlockCounter : integer; b128K : Boolean; lMemPos : integer;
  lOutFFFD : integer; sModel: string; size:integer;
begin
    lCounter := 0;
    BlockRead(hFile, sData, 2);
    lHeaderLen := 256 * sData[1] + sData[0];
    b128k := false;

    // offset 32 - 2 bytes - PC
    If lCounter < lHeaderLen Then
    begin
        BlockRead(hFile, sData, 2);
        PC.W := 256 * sData[1] + sdata[0];
        lCounter := lCounter + 2;
    End;

    // offset 34 - 1 byte - hardware mode
    If lCounter < lHeaderLen Then
    begin
        BlockRead(hFile, sData, 1);
        Case sData[0] of
          0:    // 48K spectrum
            begin
              bHardwareSupported := True;
              SetEmulatedModel(0);
            end;
          1:  //48K spectrum + Interface 1
            begin
              bHardwareSupported := True;
              SetEmulatedModel(0);
            end;
          2:  //SamROM
            begin
              sModel := 'SamROM';
              bHardwareSupported := False;
            end;
          3:
            begin
              If lHeaderLen = 23 Then
              begin
                sModel := '128K Spectrum';
                bHardwareSupported := True;
                b128K := True;
                SetEmulatedModel(1);
              end
              Else
              begin
                sModel := '48K Spectrum + M.G.T.';
                bHardwareSupported := True;
                SetEmulatedModel(0);
              End;
            end;
          4:
            begin
              If lHeaderLen = 23 Then
              begin
                sModel := '128K Spectrum + Interface 1';
                bHardwareSupported := True;
                b128K := True;
                SetEmulatedModel(1);
              end
              Else
              begin
                sModel := '128K Spectrum';
                bHardwareSupported := True;
                b128K := True;
                SetEmulatedModel(1);
              End;
            end;
          5:
            begin
              bHardwareSupported := True;
              sModel := '128K Spectrum + Interface 1';
              b128K := True;
              SetEmulatedModel(1);
            end;
          6:
            begin
              bHardwareSupported := False;
              sModel := '128K Spectrum + M.G.T.';
              b128K := True;
              SetEmulatedModel(1);
            end;
          7:
            begin
              bHardwareSupported := False;
              sModel := 'ZX Spectrum +3';
              b128K := True;
              SetEmulatedModel(2);
            end;
          Else
          begin
            bHardwareSupported := False;
            sModel := 'Unknown hardware platform';
          end;
        End;

        If bHardwareSupported = False Then
        begin
            MessageDlg(
              'The emulator does not support the required hardware platform (' + sModel + ') for this snapshot. This snapshot may not function correctly in DelphiSpec.',
              mtinformation,[mbok],0);
        End;

        lCounter := lCounter + 1;
    End;

    // offset 35 - 1 byte - last out to 0x7FFD - not required for 48K spectrum
    If lCounter < lHeaderLen Then
    begin
        If b128K Then
        begin
          BlockRead(hFile, sData, 1);
          outb($7FFD, sData[0]);
        end
        Else
        begin
            BlockRead(hFile, sData, 1);
            glPageAt[0] := nRomStart;
            glPageAt[1] := 5;
            glPageAt[2] := 1;
            glPageAt[3] := 2;
            glPageAt[4] := nRomStart;
        End;
        lCounter := lCounter + 1;
    End;

    // offset 36 - 1 byte - 0xFF if Interface 1 ROM is paged in
    If lCounter < lHeaderLen Then
    begin
        BlockRead(hFile, sData, 1);
        lCounter := lCounter + 1;

        If sData[0] = 255 Then
            messagedlg(
              'This snapshot was saved with the Interface 1 ROM paged in. It may not run correctly.',
              mtinformation,[mbok],0);
    End;

    // offset 37 - 1 byte (bit 0: 1=intR emulation on, bit 1: 1=LDIR emulation on
    If lCounter < lHeaderLen Then
    begin
        BlockRead(hFile, sData, 1);
        lCounter := lCounter + 1;
    End;

    // offset 38 - Last out to 0xFFFD (+2/+3 sound chip register number)
    lOutFFFD := 0;
    If lCounter < lHeaderLen Then
    begin
        BlockRead(hFile, sData, 1);
        lOutFFFD := sData[0];
        lCounter := lCounter + 1;
    end;

    // offset 39 - 16 bytes - contents of the sound chip registers
    If lCounter < lHeaderLen Then
    begin
        BlockRead(hFile, sData, 16);
        If b128K Then
        with ay8912 do
        begin
            Regs[0]:=sData[0];
            Regs[1]:=sData[1];
            Regs[2]:=sData[2];
            Regs[3]:=sData[3];
            Regs[4]:=sData[4];
            Regs[5]:=sData[5];
            Regs[6]:=sData[6];
            Regs[7]:=sData[7];
            Regs[8]:=sData[8];
            Regs[9]:=sData[9];
            Regs[10]:=sData[10];
            Regs[11]:=sData[11];
            Regs[12]:=sData[12];
            Regs[13]:=sData[13];
            Regs[14]:=sData[14];
            Regs[15]:=sData[15];
        end;
        lCounter := lCounter + 16;
    End;

    If b128K Then
        outb($FFFD, lOutFFFD);

    // read the remaining bytes of the header (we don't care what information they hold)
    If lCounter < lHeaderLen Then
        BlockRead(hFile, sData, lHeaderLen-lCounter);

    repeat
        // read a block
        BlockRead(hFile, sData, 2);
        If EOF(hFile) Then break;
        lHeaderLen := 256 * sData[1] + sData[0];
        BlockRead(hFile, sData, 1);
        Case sData[0] of
          0:  // Spectrum ROM
            If b128K Then lMemPage := 9 Else lMemPage := 8;
          1:  //Interface 1 ROM, or similar (we discard these blocks)
            lMemPage := -1;
          2:  //128K ROM (reset)
            If b128K Then lMemPage := 8 Else lMemPage := -1;
          3:  //Page 0 (not used by 48K Spectrum)
            If b128K Then lMemPage := 0 Else lMemPage := -1;
          4:  //Page 1 RAM at 0x8000
            lMemPage := 1;
          5:  //Page 2 RAM at 0xC000
            lMemPage := 2;
          6:  //Page 3 (not used by 48K Spectrum)
            If b128K Then lMemPage := 3 Else lMemPage := -1;
          7:  //Page 4 (not used by 48K Spectrum)
            If b128K Then lMemPage := 4 Else lMemPage := -1;
          8:  //Page 5 RAM at 0x4000
            lMemPage := 5;
          9:  //Page 6 (not used by 48K Spectrum)
            If b128K Then lMemPage := 6 Else lMemPage := -1;
          10: //Page 7 (not used by 48K Spectrum)
            If b128K Then lMemPage := 7 Else lMemPage := -1;
          11: //Multiface ROM
            lMemPage:=-1;
          else
            lMemPage:=-1;
        End;

        BlockRead(hFile, sData, lHeaderLen, size);

        If lMemPage <> -1 Then
        begin
            If lHeaderLen = $FFFF Then
            begin
                //Not a compressed block, just copy it straight into RAM
                For lCounter := 0 To 16383 do
                    gRAMPage[lMemPage, lCounter] := sData[lCounter];
            end
            Else
            begin
                // Uncompress the block to memory
                lCounter := 0;
                lMemPos := 0;
                repeat
                    If sData[lCounter] = $ED Then
                    begin
                        If sData[lCounter + 1] = $ED Then
                        begin
                            // This is an encoded block
                            lCounter := lCounter + 2;
                            lHeaderLen := sData[lCounter];
                            lCounter := lCounter + 1;
                            For lBlockCounter := 0 To lHeaderLen - 1 do
                                gRAMPage[lMemPage, lMemPos + lBlockCounter] := sData[lCounter];

                            lMemPos := lMemPos + lHeaderlen;
                        end
                        Else
                        begin
                            // Just a single ED, write it out
                            gRAMPage[lMemPage, lMemPos] := $ED;
                            lMemPos := lMemPos + 1;
                        End;
                    end
                    Else
                    begin
                        gRAMPage[lMemPage, lMemPos] := sData[lCounter];
                        inc(lMemPos);
                    End;
                    inc(lCounter);
                until lCounter >= size;
            End;
        End;
    Until Eof(hFile);
End;


procedure ReadZ80V1Snap(var hFile : File);
    var
    lDataLen : integer; sData : array[0..65535] of Byte; lBlockLen : integer;
    lCounter : integer; lMemPos : integer; lBlockCounter : integer;
    size : integer;
begin
    lDataLen := FileSize(hFile)-FilePos(hFile);

    //read the compressed data into sData
    BlockRead(hFile, sData, lDataLen, size);

    // Uncompress the block to memory
    lCounter := 0;
    lMemPos := 16384;
    repeat
        If sData[lCounter] = $ED Then
        begin
            If sData[lCounter + 1] = $ED Then
            begin
                // This is an encoded block
                lCounter := lCounter + 2;
                lBlockLen := sData[lCounter];
                inc(lCounter);
                For lBlockCounter := 1 To lBlockLen do
                begin
                    gRAMPage[glPageAt[lMemPos div 16384], lMemPos And 16383] := sData[lCounter];
                    inc(lMemPos);
                end
            end
            Else
            begin
                // Just a single ED, write it out
                gRAMPage[glPageAt[lMemPos div 16384], lMemPos And 16383] := $ED;
                inc(lMemPos);
            End
        end
        Else
        begin
            gRAMPage[glPageAt[lMemPos div 16384], lMemPos And 16383] := sData[lCounter];
            inc(lMemPos);
        End;
        inc(lCounter);
    Until lCounter > size - 4;

    If (sData[lCounter-1] <> 0)
      or (sData[lCounter] <> $ED)
      or (sData[lCounter+1] <> $ED)
      or (sData[lCounter+2] <> 0) Then
        MessageDLG(
              'Error in compressed Z80 file. Block end marker 0x00EDED00 is not present.',
              mterror,[mbok],0)

End;

procedure LoadZ80Snap(sFilename : String);
    var
      hFile: File; sData : array[0..65535] of byte; iCounter : integer;
      bCompressed : Boolean;
      numread:integer;
begin
    AssignFile(hFile, sFileName);
    filemode:=0;
    reset(hFile, 1);

{$IFDEF USE_DEBUGGING}
    ReportDebug(0,'LoadZ80_0',0,True);
{$ENDIF}
    glPageAt[0] := nRomStart;
    glPageAt[1] := 5;
    glPageAt[2] := 2;
    glPageAt[3] := 0;
    glPageAt[4] := nRomStart;

    {$IFDEF USE_SOUND}
    If gbSoundEnabled Then AY8912.reset;
    {$ENDIF}

    //byte 0 - A register
    BlockRead(hFile, sData, 1);
    AF.h := sData[0];

    //byte 1 - F register
    BlockRead(hFile, sData, 1);
    AF.l:=sData[0];

    //bytes 2 + 3 - BC register pair (C first, then B)
    BlockRead(hFile, sData, 2);
    BC.W:=256*sData[1]+sData[0];

    //bytes 4 + 5 - HL register pair
    BlockRead(hFile, sData, 2);
    HL.W := 256*sData[1]+sData[0];

    //bytes 6 + 7 - PC (this is zero for v2.x or v3.0 Z80 files)
    BlockRead(hFile, sData, 2);
    PC.W := 256*sData[1]+sData[0];

    //bytes 8 + 9 - SP
    BlockRead(hFile, sData, 2);
    SP.w := 256*sData[1]+sData[0];

    //byte 10 - Interrupt register
    BlockRead(hFile, sData, 1);
    IR.h := sData[0];

    //byte 11 - Refresh register
    BlockRead(hFile, sData, 1);
    IR.l := (sData[0] and 127);

    //byte 12 - bitfield
    BlockRead(hFile, sData, 1);
    //if byte 12 = 255 then it must be treated as if it = 1, for compatibility with other emulators
    If sData[0] = 255 Then sData[0] := 1;
    //bit 0 - bit 7 of R
    If (sData[0] And 1) = 1 Then IR.l := iR.l Or 128;
    RTemp := IR.l;

    //bits 1,2 and 3 - border color
    glNewBorder := (sData[0] And 14) div 2;
    //bit 4 - 1 if SamROM switched in (we don't care about this!)
    //bit 5 - if 1 and PC<>0 then the snapshot is compressed using the
    //         rudimentary Z80 run-length encoding scheme
    bCompressed:=false;
    If (sData[0] And $20) = $20 Then bCompressed := True;
    //bits 6 + 7 - no meaning

    //bytes 13 + 14 - DE register pair
    BlockRead(hFile, sData, 2);
    DE.W := 256*sData[1]+sData[0];

    //bytes 15 + 16 - BC' register pair
    BlockRead(hFile, sData, 2);
    BC2.W := 256*sData[1]+sData[0];

    //bytes 17 + 18 - DE' register pair
    BlockRead(hFile, sData, 2);
    DE2.W := 256*sData[1]+sData[0];

    //bytes 19 + 20 - HL' register pair
    BlockRead(hFile, sData, 2);
    HL2.W := 256*sData[1]+sData[0];

    //bytes 21 + 22 - AF' register pair (A first then F - not Z80 byte order!!)
    BlockRead(hFile, sData, 2);
    AF2.W := 256*sData[0]+sData[1];

    //byte 23 + 24 - IY register pair
    BlockRead(hFile, sData, 2);
    IY.W := 256*sData[1]+sData[0];

    //byte 25 + 26 - IX register pair
    BlockRead(hFile, sData, 2);
    ix.w := 256*sData[1]+sData[0];

    //byte 27 - Interrupt flipflop (0=DI, else EI)
    BlockRead(hFile, sData, 1);
    If sData[0] = 0 Then
    begin
        IFF1 := 0;
        IFF2 := 0;
    end
    Else
    begin
        IFF1 := 1;
        IFF2 := 1;
    End;
    //byte 28 - IFF2 (ignored)
    BlockRead(hFile, sData, 1);

    //byte 29 - Interrupt mode (bits 2 - 7 contain info about joystick modes etc, which we ignore)
    BlockRead(hFile, sData, 1);
    IM := sData[0] And 3;
{$IFDEF USE_DEBUGGING}
    ReportDebug(0,'LoadZ80_1',0,True);
{$ENDIF}

    If PC.W = 0 Then
    begin
        //This is a V2 or V3 Z80 file
        ReadZ80V2orV3Snap(hFile);
        CloseFile(hFile);
    end
    Else
    begin
        SetEmulatedModel(0); // V1 .Z80 snapshots are all 48K
        // PC<>0, so lets check to see if this is a compressed V1 Z80 file
        If bCompressed Then
        begin
            //Uncompress the RAM data
            ReadZ80V1Snap(hFile);
            CloseFile(hFile);
        end
        Else
        begin
            // Uncompressed Z80 file
            BlockRead(hFile, sData, 49152,numread);
            CloseFile(hFile);

            // Copy the RAM data to addressable memory space
            For iCounter := 16384 To 65535 do
              gRAMPage[glPageAt[iCounter div 16384], iCounter And 16383] := sData[iCounter - 16384];
        End;
    End;
{$IFDEF USE_DEBUGGING}
    ReportDebug(0,'LoadZ80_2',0,True);
{$ENDIF}

    // Set the initial border color
    display.border:= (glNewBorder and 7);

    screenPaint;
    resetKeyboard;
End;




Procedure SaveROM(sFilename : String);
  var hFile : File; lCounter : integer;
begin
  AssignFile(hFile, sFileName);
  filemode:=2;
  rewrite(hFile, 1);
  for lCounter := 0 to 16383 do
    BlockWrite(hFile, gRAMPage[glPageAt[0], lCounter], 1);

  CloseFile(hFile);
End;

procedure SaveSNA128Snap(var hFile : File);
  var lBank : integer; lCounter : integer;
  var bDummy: byte; wDummy: word;
begin
  BlockWrite(hFile, IR.h, 1);
  BlockWrite(hFile, HL2.W, 2);
  BlockWrite(hFile, DE2.W, 2);
  BlockWrite(hFile, BC2.W, 2);
  BlockWrite(hFile, AF2.W, 2);

  BlockWrite(hFile, HL.W, 2);
  BlockWrite(hFile, DE.W, 2);
  BlockWrite(hFile, BC.W, 2);
  BlockWrite(hFile, IY.W, 2);
  BlockWrite(hFile, ix.w, 2);

  //Interrupt flipflops
  If IFF1 = 1 Then
    bdummy := 4
  Else
    bdummy := 0;
  BlockWrite(hFile, bdummy, 1);

  //R
  RTemp := RTemp And 127;
  bdummy := (IR.L And $80) Or RTemp;
  BlockWrite(hFile, bdummy, 1);

  // AF
  BlockWrite(hFile, AF.W, 2);

  // SP
  BlockWrite(hFile, SP.W, 2);

  // Interrupt Mode
  BlockWrite(hFile, IM, 1);

  bDummy := display.Border;
  BlockWrite(hFile, bDummy, 1);

  // Save the three currently-paged RAM banks
  for lCounter:=1 to 3 do
   BlockWrite(hFile, gRAMPage[glPageAt[lCounter]],16384);

  //PC
  BlockWrite(hFile, PC.W, 2);

  //Last out to 0x7FFD
  BlockWrite(hFile, glLastOut7FFD, 1);

  //Is TR-DOS paged? (0=not paged, 1=paged)
  bDummy := byte(glEmulatedTRDOS);
  BlockWrite(hFile, bDummy, 1);

  // Save the remaining RAM banks
  lBank := 0;
  While lBank < 8 do
  begin
    If (lBank <> glPageAt[1]) And (lBank <> glPageAt[2]) And (lBank <> glPageAt[3]) Then
     BlockWrite(hFile, gRAMPage[lbank],16384);
    inc(lBank);
  end;
End;

procedure SaveSNASnap(sFilename : String);
    var hFile : File; lCounter : integer; bDummy: byte; wDummy: word;
begin

  AssignFile(hFile, sFileName);
  filemode:=2;
  rewrite(hFile, 1);

  If glEmulatedModel <> 0 Then
  begin
      // We're not running in 48K mode
      SaveSNA128Snap(hFile);
      CloseFile(hFile);
      Exit;
  end;
  push(pc.w);

  BlockWrite(hFile, IR.H, 1);
  BlockWrite(hFile, HL2.W, 2);
  BlockWrite(hFile, DE2.W, 2);
  BlockWrite(hFile, BC2.W, 2);
  BlockWrite(hFile, AF2.W, 2);

  BlockWrite(hFile, HL.W, 2);
  BlockWrite(hFile, DE.W, 2);
  BlockWrite(hFile, BC.W, 2);
  BlockWrite(hFile, IY.W, 2);
  BlockWrite(hFile, ix.w, 2);

  //Interrupt flipflops
  If IFF1 = 1 Then
    bDummy := 4
  Else
    bDummy := 0;
  BlockWrite(hFile, bDummy, 1);

  //R
  RTemp := RTemp And 127;
  bDummy := (IR.l And $80) Or RTemp;
  BlockWrite(hFile, bDummy, 1);

  //AF
  BlockWrite(hFile, AF.W, 2);

  //SP
  BlockWrite(hFile, SP.W, 2);

  //Interrupt Mode
  BlockWrite(hFile, IM, 1);

  BlockWrite(hFile, display.border, 1);

  For lCounter := 0 To 16383 do
      BlockWrite(hFile, gRAMPage[glPageAt[1], lCounter], 1);
  For lCounter := 0 To 16383 do
      BlockWrite(hFile, gRAMPage[glPageAt[2], lCounter], 1);
  For lCounter := 0 To 16383 do
      BlockWrite(hFile, gRAMPage[glPageAt[3], lCounter], 1);

  CloseFile(hFile);
  pop(pc.w);
End;

Function CompressMemoryBlock(lBlock : integer; pData : pByte) : integer;
    var pDataStart: pByte; ii: integer; lCounter : integer; lCounter2 : integer; lByteCount : integer;
begin
    pDataStart := pData;
    lCounter := 0;
    repeat
        If gRAMPage[lBlock, lCounter] = gRAMPage[lBlock, lCounter + 1] Then
        begin
          // String of two or more bytes starting from lCounter
          // Might be a candidate for compression
          lCounter2 := lCounter;
          While (lCounter2 - lCounter) < 260 do
          begin
              if (lCounter2 >= 16383) then
                break;
                
              If gRAMPage[lBlock, lCounter2 + 1] = gRAMPage[lBlock, lCounter2] Then
                  lCounter2 := lCounter2 + 1
              Else
                  break;
          end;
          lByteCount := lCounter2 - lCounter + 1;
          lCounter := lCounter2 + 1;
          If lByteCount > 255 Then
          begin
              lCounter := lCounter - (lByteCount - 255);
              lByteCount := 255;
          End;

          If lByteCount > 4 Then
          begin
              pData^ := $ED;
              inc(pData);
              pData^ := $ED;
              inc(pData);
              pData^ := lByteCount;
              inc(pData);
              pData^ := gRAMPage[lBlock, lCounter - 1];
              inc(pData);
              If lCounter < 16384 Then
              begin
                  pData^ := gRAMPage[lBlock, lCounter];
                  inc(pData);
                  lCounter := lCounter + 1;
              End;
          end
          Else
          begin
              If gRAMPage[lBlock, lCounter - 1] = $ED Then
              begin
                  // Strings of 2 or more ED bytes are always encoded
                  pData^ := $ED;
                  inc(pData);
                  pData^ := $ED;
                  inc(pData);
                  pData^ := lByteCount;
                  inc(pData);
                  pData^ := gRAMPage[lBlock, lCounter - 1];
                  inc(pData);
                  If lCounter < 16384 Then
                  begin
                    pData^ := gRAMPage[lBlock, lCounter];
                    inc(pData);
                    lCounter := lCounter + 1;
                  End;
              end
              Else
              begin
                  // Strings of 4 or fewer bytes are not encoded
                  for ii := 1 to lByteCount do
                  begin
                    pData^ := gRAMPage[lBlock, lCounter - 1];
                    inc(pData);
                  end;
              end;
          End;
        end
        Else
        begin
          // A single byte, just write it out
          pData^ := gRAMPage[lBlock, lCounter];
          inc(pData);
          lCounter := lCounter + 1;
        End;
    Until lCounter >= 16383;

    If lCounter = 16383 Then
    begin;
        // Write out the final byte
        pData^ := gRAMPage[lBlock, 16383];
        inc(pData);
    end;

    lCounter2 := Integer(pData);
    lCounter := Integer(pDataStart);
    CompressMemoryBlock := lCounter2-lCounter;
End;

procedure SaveZ80Snap(sFilename : String);
    var hFile : File; sData : array[0..65535] of byte;
      lCounter : integer; lBufSize : integer; bDummy:byte;
begin
  AssignFile(hFile, sFileName);
  filemode:=2;
  rewrite(hFile, 1);

  //Index0:A,F
  BlockWrite(hFile, AF.w, 2);

  //Index2:BC
  BlockWrite(hFile, BC.w,2);

  //Index4:HL
  BlockWrite(hFile, HL.W, 2);

  //Index6
  //Set PC to zero to indicate a v3.0 Z80 file
  bDummy := 0;
  BlockWrite(hFile, bDummy, 1);
  BlockWrite(hFile, bDummy, 1);

  //Index8
  //SP
  BlockWrite(hFile, SP.W, 2);

  //Index10
  //I
  BlockWrite(hFile, IR.h, 1);

  //Index11
  //R (7 bits)
  RTemp := RTemp And $7F;
  BlockWrite(hFile, RTemp, 1);

  //Index12
  //bitfield
  if IR.l and $80 = $80 then
    bDummy := 1
  else
    bDummy := 0;
  bDummy := bDummy or (display.border* 2);
  BlockWrite(hFile, bDummy, 1);

  //Index13
  //DE
  BlockWrite(hFile, DE.W, 2);

  //Index15
  //BC'
  BlockWrite(hFile, BC2.W, 2);

  //Index17
  //DE'
  BlockWrite(hFile, DE2.W, 2);

  //Index19
  //HL'
  BlockWrite(hFile, HL2.W, 2);

  //Index21
  //AF'
  bDummy := AF2.W div 256;
  BlockWrite(hFile, bDummy, 1);
  bDummy := AF2.W and 255;
  BlockWrite(hFile, bDummy, 1);

  //Index23
  //IY
  BlockWrite(hFile, IY.w, 2);

  //Index25
  //IX
  BlockWrite(hFile, ix.w, 2);

  //Index27
  //Interrupt flipflops
  If IFF1 = 1 Then
  begin
    bDummy := 255;
    bDummy := 255;
  end
  Else
  begin
    bDummy := 0;
    bDummy := 0;
  End;
  BlockWrite(hFile, bDummy, 1);
  BlockWrite(hFile, bDummy, 1);

  //Index29
  // Interrupt Mode
  BlockWrite(hFile, IM, 1);

  //Index30
  // V2.01 info
  bDummy := 23; BlockWrite(hFile, bDummy, 1);
  bDummy := 0; BlockWrite(hFile, bDummy, 1);

  //Index32
  //PC
  BlockWrite(hFile, PC.W, 2);

  //Index34
  //Hardware mode
  If glEmulatedModel = 0 Then
  begin
      bDummy := 0;    //48K Spectrum
      BlockWrite(hFile, bDummy, 1);
      BlockWrite(hFile, bDummy, 1);
  end
  Else
  begin
      bDummy := 3;    //128K Spectrum
      BlockWrite(hFile, bDummy, 1);

      //Last out to 7FFD
      BlockWrite(hFile, glLastOut7FFD, 1);
  End;

  //Index36
  //IF.1 Paged in
  bDummy := 0; BlockWrite(hFile, bDummy, 1);

  //Index37
  //1=R emulation on, 2=LDIR emulation on
  bDummy := 3; BlockWrite(hFile, bDummy, 1);

  //Index38
  {$IFDEF USE_SOUND}
  bDummy := glSoundRegister; BlockWrite(hFile, bDummy, 1);
  {$ELSE}
  bDummy := 0; BlockWrite(hFile, bDummy, 1);
  {$ENDIF}

  //Index39
  //AY-3-8912 register contents
  For lCounter := 0 To 15 do
   begin
    bdummy:=AY8912.Regs[lCounter];
    BlockWrite(hFile, bdummy, 1);
   end;

  //Index54
  If glEmulatedModel = 0 Then
  begin
    // Block 1
    lBufSize := CompressMemoryBlock(glPageAt[1], Addr(sData));


    // Buffer length
    BlockWrite(hFile, lBufSize, 2);


    //Index49
    //Block number
    bDummy := glPageAt[1] + 3;
    BlockWrite(hFile, bDummy, 1);
    BlockWrite(hFile, sData, lBufSize);

    //Block 2
    lBufSize := CompressMemoryBlock(glPageAt[2], Addr(sData));
    //Buffer length
    BlockWrite(hFile, lBufSize, 2);
    //Block number
    bDummy := glPageAt[2] + 3;
    BlockWrite(hFile, bDummy, 1);
    BlockWrite(hFile, sData, lBufSize);

    //Block 3
    lBufSize := CompressMemoryBlock(glPageAt[3], Addr(sData));
    //Buffer length
    BlockWrite(hFile, lBufSize, 2);
    //Block number
    bDummy := glPageAt[3] + 3;
    BlockWrite(hFile, bDummy, 1);
    BlockWrite(hFile, sData, lBufSize);
  end
  Else
  begin
      For lCounter := 0 To 7 do
      begin
        lBufSize := CompressMemoryBlock(lCounter, Addr(sData));
        //Buffer length
        BlockWrite(hFile, lBufSize, 2);
        //Block number
        bDummy := lCounter + 3;
        BlockWrite(hFile, bDummy, 1);
        BlockWrite(hFile, sData, lBufSize);
      end;
  End;

  CloseFile(hFile);
End;


end.
