unit UMain;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, StdCtrls, Buttons, Menus, ComCtrls, ModDisplay, ImgList,
  UResolution,ToolWin, AgOpenDialog;

type
  TfrmMainWnd = class(TForm)
    mnuMain: TMainMenu;
    mnuFile: TMenuItem;
    mnuOptions: TMenuItem;
    mnuHelp: TMenuItem;
    mnuFileOpen: TMenuItem;
    mnuFileSaveAs: TMenuItem;
    N1: TMenuItem;
    mnuFileResetSpectrum: TMenuItem;
    N2: TMenuItem;
    mnuFileExit: TMenuItem;
    mnuOptionsGeneralSettings: TMenuItem;
    mnuOptionsDisplay: TMenuItem;
    mnuHelpSpectrumKeyboard: TMenuItem;
    mnuHelpAbout: TMenuItem;
    dlgOpen: TAgOpenDialog;
    dlgSave: TAgSaveDialog;
    mnuNMI: TMenuItem;
    Toolbar: TToolBar;
    tbReset: TToolButton;
    ImageList1: TImageList;
    ToolButton2: TToolButton;
    tbOpen: TToolButton;
    tbSave: TToolButton;
    ToolButton7: TToolButton;
    tbSpec48: TToolButton;
    tbSpec128: TToolButton;
    tbPentagon: TToolButton;
    ToolButton9: TToolButton;
    tbSetup: TToolButton;
    tbKeyboard: TToolButton;
    tbDebug: TToolButton;
    ToolButton13: TToolButton;
    tbPause: TToolButton;
    StatusBar: TStatusBar;
    mnuInsertDisk: TMenuItem;
    mnuTapebrowser: TMenuItem;
    View1: TMenuItem;
    mnuConsole: TMenuItem;
    mnuDebugger: TMenuItem;
    N3: TMenuItem;
    mnuFullscreen: TMenuItem;
    procedure MyIdleHandler(Sender: TObject; var Done: Boolean);
    procedure ContinueEmulation(Sender: TObject);
    procedure StopEmulation(Sender: TObject);
    procedure mnuFileOpenClick(Sender: TObject);
    procedure mnuFileExitClick(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure FormKeyUp(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure mnuFileResetSpectrumClick(Sender: TObject);
    procedure mnuOptionsGeneralSettingsClick(Sender: TObject);
    procedure mnuOptionsDisplayClick(Sender: TObject);
    procedure mnuHelpSpectrumKeyboardClick(Sender: TObject);
    procedure mnuHelpAboutClick(Sender: TObject);
    procedure mnuFileSaveAsClick(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormActivate(Sender: TObject);
    procedure mnuNMIClick(Sender: TObject);
    procedure tbResetClick(Sender: TObject);
    procedure tbOpenClick(Sender: TObject);
    procedure tbSaveClick(Sender: TObject);
    procedure tbModelClick(Sender: TObject);
    procedure tbSetupClick(Sender: TObject);
    procedure tbKeyboardClick(Sender: TObject);
    procedure tbPauseClick(Sender: TObject);
    procedure tbDebugClick(Sender: TObject);
    procedure mnuInsertDiskClick(Sender: TObject);
    procedure FormMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure mnuTapebrowserClick(Sender: TObject);
    procedure mnuDebuggerClick(Sender: TObject);
    procedure mnuConsoleClick(Sender: TObject);
    procedure FormPaint(Sender: TObject);
    procedure mnuFullscreenClick(Sender: TObject);
  private
    ffullwindow:Boolean;
    ffullscreen:Boolean;
    fdiff:integer;
    procedure ExecConfig;
    procedure ShowCaption;
    procedure RemoveCaption;
    procedure SetWindow;
    procedure WMEraseBkgnd(var m: TWMEraseBkgnd); message WM_ERASEBKGND;
    procedure doKey(down : Boolean; ascii : Word; mods : TShiftState);
    procedure FileOpenDialog;
    procedure FileSaveAsDialog;
    procedure ResetSound;
    { Private declarations }
  public
    { Public declarations }
    property fullwindow:Boolean read ffullwindow;
  end;

var
  frmMainWnd: TfrmMainWnd;
  Global_TStates: integer;
  z80runstate: integer;

{$IFDEF USE_DEBUGGING}
 procedure ReportDebug(MSGType: Integer;Desc : string; NextPc: integer; forceOutput: boolean);
{$ENDIF}
procedure EvalDebug;
procedure EvalBreakpoints;

implementation

uses modMain, ModTape,modz80, modAY8912, modwaveout,
     modloader,modBetadisk,uDisasm,
     mmsystem,
     UKeyboard, UAbout, UDebug, USetup, UConsole;

{$R *.DFM}

var ReportDebugCallCtr:integer;

procedure ReportDebug(MSGType: Integer;Desc : string; NextPc: integer; forceOutput: boolean);
var ii:integer;
    S,Msg:String;
begin
 with frmconsole.mmoconsole.lines do
  begin
    if (ReportDebugCallCtr >= 0) and (ReportDebugCallCtr <= 1000000) then
    begin
      if ((ReportDebugCallCtr mod 100000) = 0) or forceOutput=True then
      begin
        case msgtype of
         0: Msg:='Info:         ';
         1: Msg:='Warning:      ';
         else Msg:='Error:        ';
        end;
        add(Format('%s%s'#9'PC=%x'#9'Tstates=%d',[Msg, Desc, NextPc, Global_TStates]));
        for ii:=1 to length(Msg) do Msg[ii]:=' ';

        if Msgtype<>0 then
        begin
         add( Format('%s %s:%x;%x;%x;%x;%x;%x;%x;%x;%x',
               [Msg,'A-IX',AF.H,HL.W,BC.W,DE.W,AF2.W,HL.W,BC.W,DE.W,IX.W]));

         add( Format('%s %s:%x;%x;%x;%x;%x;%x;%x;%x;%x;%x',
           [Msg,'IY-Flags', IY.W,IX.W,SP.W,PC.W,IR.H,IR.L, byte(Iff1),byte(Iff2),IM,AF.l]));

         s:='';
         For ii := low(gRAMPage) To high(gRAMPage) do
           s:=s+'0x'+inttohex(ramCrc(ii),4)+';';
         add(Msg+'RAM Crc: '+s);
        end;
      end;
    end;
   enD;
   Inc(ReportDebugCallCtr);
end;

procedure EvalDebug;
const aBit: array[1..4] of byte = ($80, $40, $20, $10);
var
  aROMInstr: array[0..3] of Byte;
  lfound:Boolean;
  I, J, nIndex: Integer;
  cInstr, cAux:String;

begin
 with frmdebug do
 begin
  if not frmdebug.visible then exit;
  lblAF.Caption := IntToHex(AF.W, 4);
  lblBC.Caption := IntToHex(BC.W, 4);
  lblDE.Caption := IntToHex(DE.W, 4);
  lblHL.Caption := IntToHex(HL.W, 4);
  lblIX.Caption := IntToHex(IX.W, 4);
  lblIY.Caption := IntToHex(IY.W, 4);
  lblSP.Caption := IntToHex(SP.W, 4);
  lblPC.Caption := IntToHex(PC.W, 4);
  lblIR.Caption := IntToHex(IR.W, 4);
  lblIFF1.Caption := IntToHex(IFF1, 1);
  lblIFF2.Caption := IntToHex(IFF1, 1);

  lblRam0.caption := inttohex(glpageat[0],2);
  lblRam1.caption := inttohex(glpageat[1],2);
  lblRam2.caption := inttohex(glpageat[2],2);
  lblRam3.caption := inttohex(glpageat[3],2);
  lblRam4.caption := inttohex(glpageat[4],2);
  lblVideo.caption := inttohex(glUseScreen,2);

  lblRam0.Font.Color:=clNavy;lblRam1.Font.Color:=clNavy;
  lblRam2.Font.Color:=clNavy;lblRam3.Font.Color:=clNavy;
  lblRam4.Font.Color:=clNavy;

  case (pc.w and $C000) shr 14 of
   0: lblRam0.Font.Color:=clMaroon;
   1: lblRam1.Font.Color:=clMaroon;
   2: lblRam2.Font.Color:=clMaroon;
   3: lblRam3.Font.Color:=clMaroon;
   4: lblRam4.Font.Color:=clMaroon;
  end;

  case IM of
    0: lblIM.Caption := '0';
    2: lblIM.Caption := '1';
    3: lblIM.Caption := '2';
    else lblIM.Caption := 'undc';
  end;
  lblSFlag.Visible := (AF.l and $80)<>0;
  lblZFlag.Visible := (AF.l and $40)<>0;
  lblHFlag.Visible := (AF.l and $10)<>0;
  lblPFlag.Visible := (AF.l and $04)<>0;
  lblNFlag.Visible := (AF.l and $02)<>0;
  lblCFlag.Visible := (AF.l and $01)<>0;
  lblBit5Flag.Visible := (AF.l and $20)<>0;
  lblBit3Flag.Visible := (AF.l and $08)<>0;
  EvalWatches;

  aROMInstr[0] := peekb(PC.W+0);
  aROMInstr[1] := peekb(PC.W+1);
  aROMInstr[2] := peekb(PC.W+2);
  aROMInstr[3] := peekb(PC.W+3);
  lFound := false;
  cAux := '';

  for i := 1 to MAX_Z80_INSTR do
  begin
    with aZ80Instr[i] do begin
      lFound := true;
      for j := 1 to nMask and $0F do
        if (nMask and aBit[j])<>0 then begin
          if aROMInstr[j-1]<>aInstr[j-1] then lFound := false;
        end
        else begin
          cAux := IntToHex(aROMInstr[j-1], 2)+cAux;
        end;
      nIndex := i;
      if lFound then begin
        break;
      end;
    end;
  end;

  if lFound then
  begin
    cInstr := aZ80Instr[nIndex].cInstr;
    i := Pos('####', cInstr);
    if i<>0 then begin
      cInstr[i] := cAux[1];
      cInstr[i+1] := cAux[2];
      cInstr[i+2] := cAux[3];
      cInstr[i+3] := cAux[4];
    end;

    i := Pos('##', cInstr);
    if i<>0 then begin
      cInstr[i] := cAux[1];
      cInstr[i+1] := cAux[2];
    end;

    cAux := '';
    for j := 1 to aZ80Instr[nIndex].nMask and $0F do
      cAux := cAux+IntToHex(aROMInstr[j-1], 2)+' ';

    lblOper.Caption := cInstr;
    lblBytes.Caption := cAux;
  end
  else begin
    lblOper.Caption := '?????????';
    lblBytes.Caption := '?? ?? ?? ??';
  end;
 end;
end;

procedure EvalBreakPoints;
var
 nBrk: Integer;

begin
 with frmdebug do
  if cbBreakpoint.checked then
  begin
   for nBrk := 0 to Breakcount-1 do
   if BreakPoints[nBrk]=pc.w then
   begin
    EvalDebug;
    DebugState:=STATE_PAUSED;
    ShowMessage('The instruction in address of the breakpoint ['+inttohex(pc.w,4)+'] is the next to be executed.');
   end;
  end;
end;

procedure TfrmMainWnd.RemoveCaption;
begin
  FDiff := GetSystemMetrics(SM_CYCAPTION);
  SetWindowLong(Handle, GWL_STYLE, GetWindowLong(Handle, GWL_Style) and not WS_Caption);
  Height:=Height - FDiff;
  Height:=Height + FDiff;
end;

procedure TfrmMainWnd.ShowCaption;
begin
  If Self.ComponentState = [csDestroying] Then exit;
  SetWindowLong(Handle, GWL_STYLE, GetWindowLong(Handle, GWL_Style) + WS_Caption);
  Height:=Height + FDiff;
  Height:=Height - FDiff;
end;

procedure TfrmMainWnd.SetWindow;
begin
 self.Hide;
 ffullwindow:=not ffullwindow;
 if ffullwindow then
  begin
   Menu:=nil;
   toolbar.hide;
   statusbar.Hide;
   RemoveCaption;
  end else
  begin
   Menu:=mnuMain;
   toolbar.show;
   statusbar.show;
   ShowCaption;
  end;
 SetDisplaySize(glDisplayWidth, glDisplayHeight);
 self.show;
 repaint;
end;

//Don't let Windows erase the background of the window
//because we fill everything ourselves
procedure TfrmMainWnd.WMEraseBkgnd(var m : TWMEraseBkgnd);
begin
  m.Result := LRESULT(False);
end;


//It runs spectrum for one interrupt period (50 ms spectrum time)
//and tells then Windows it needs more processor cycles (Done:=False).
//This way it is a definite CPU hog, but should be fast too
//During setting screens, emulation is paused (z80runstate=Z80PAUSED)

procedure TfrmMainWnd.MyIdleHandler(Sender: TObject; var Done: Boolean);
var dummy:byte;
begin
 if glEmulatedTRDos then betadisk.diskVG($A,dummy);
  if (z80runstate=Z80RUNNING) then
  begin
    //Execute some z80 code
    execute(Global_TStates);
    EvalDebug;
    Done := false;
  end
  else
  begin
    Sleep(100);
    Done := true;
  end;
 if glEmulatedTRDos then betadisk.diskVG($B,dummy);
end;

//When app gains focus
procedure TfrmMainWnd.ContinueEmulation(Sender: TObject);
begin;
  if z80runstate=Z80STOPPED then
    z80runstate := Z80RUNNING;
end;


//When app loses focus
procedure TfrmMainWnd.StopEmulation(Sender: TObject);
begin;
  if z80runstate=Z80RUNNING then
    z80runstate := Z80STOPPED;
end;


procedure TfrmMainWnd.FileOpenDialog;
  var Ext: string;
begin
    if (dlgOpen.execute) then
    begin
      If dlgOpen.FileName <> '' Then
      begin
        Ext := LowerCase(ExtractFileExt(dlgOpen.FileName));

        if (Ext = '.tap') then tape.tapename:=dlgOpen.Filename else
        if (Ext = '.trd') or (Ext='.scl') or (Pos('$',ext)>0) then betadisk.disks[0].filename:=dlgOpen.Filename else
        if (Ext = '.z80') then LoadZ80Snap(dlgOpen.FileName) else
        if (Ext = '.rom') then
         begin
           Z80Reset;
           SetEmulatedModel(0);
           if LoadROM(dlgOpen.FileName) = false then
             Application.Terminate;
         end
         else if (Ext = '.sna') then
           LoadSNASnap(dlgOpen.FileName)
         else
        begin
          //try opening it as a SNA file
          LoadSNASnap(dlgOpen.FileName);
        end;
      end;
    end;
End;


procedure TfrmMainWnd.mnuFileOpenClick(Sender: TObject);
  var ii: integer;
begin
  {$IFDEF USE_SOUND}
      If gbSoundEnabled Then waveOutReset(glphWaveOut);
  {$ENDIF}

  FileOpenDialog;
  {$IFDEF USE_SOUND}
      If gbSoundEnabled Then
      begin
          waveOutReset(glphWaveOut);
          For ii := 1 To 12 do
          begin
              waveOutWrite(glphWaveOut,
                @gtWavHdr[ii], sizeof(gtWavHdr[ii]));
          end;
          glBufNum := 12;
      End;
  {$ENDIF}
end;

procedure TfrmMainWnd.mnuFileExitClick(Sender: TObject);
begin
    {$IFDEF USE_SOUND}
    If gbSoundEnabled Then CloseWaveOut;
    {$ENDIF}

    timeEndPeriod(1);
    Application.Terminate;
end;

procedure TfrmMainWnd.doKey(down : Boolean; ascii : Word; mods : TShiftState);
    var CAPS : Boolean; SYMB : Boolean;
    //Shift : Boolean;
begin

    CAPS := (ssShift in mods);
    SYMB := (ssCtrl in mods);

    // Change control versions of keys to lower case
    If (ascii >= 1) And (ascii <= $27) And SYMB Then
        ascii := ascii + Ord('a') - 1;

    If CAPS Then keyCAPS_V := (keyCAPS_V And (Not 1)) Else keyCAPS_V := (keyCAPS_V Or 1);
    If SYMB Then keyB_SPC := (keyB_SPC And (Not 2)) Else keyB_SPC := (keyB_SPC Or 2);

    Case ascii of
      8:  //Backspace
        begin
          If down Then
          begin
            key6_0 := (key6_0 And (Not 1));
            keyCAPS_V := (keyCAPS_V And (Not 1));
          end
          Else
          begin
            key6_0 := (key6_0 Or 1);
            If Not CAPS Then
                keyCAPS_V := (keyCAPS_V Or 1);
          End;
        end;
      65: //A
        If down Then keyA_G := (keyA_G And (Not 1)) Else keyA_G := (keyA_G Or 1);
      66: //B
        If down Then keyB_SPC := (keyB_SPC And (Not 16)) Else keyB_SPC := (keyB_SPC Or 16);
      67: //C
        If down Then keyCAPS_V := (keyCAPS_V And (Not 8)) Else keyCAPS_V := (keyCAPS_V Or 8);
      68: //D
        If down Then keyA_G := (keyA_G And (Not 4)) Else keyA_G := (keyA_G Or 4);
      69: //E
        If down Then keyQ_T := (keyQ_T And (Not 4)) Else keyQ_T := (keyQ_T Or 4);
      70: //F
        If down Then keyA_G := (keyA_G And (Not 8)) Else keyA_G := (keyA_G Or 8);
      71: //G
        If down Then keyA_G := (keyA_G And (Not 16)) Else keyA_G := (keyA_G Or 16);
      72: //H
        If down Then keyH_ENT := (keyH_ENT And (Not 16)) Else keyH_ENT := (keyH_ENT Or 16);
      73: //I
        If down Then keyY_P := (keyY_P And (Not 4)) Else keyY_P := (keyH_ENT Or 4);
      74: //J
        If down Then keyH_ENT := (keyH_ENT And (Not 8)) Else keyH_ENT := (keyH_ENT Or 8);
      75: //K
        If down Then keyH_ENT := (keyH_ENT And (Not 4)) Else keyH_ENT := (keyH_ENT Or 4);
      76: //L
        If down Then keyH_ENT := (keyH_ENT And (Not 2)) Else keyH_ENT := (keyH_ENT Or 2);
      77: //M
        If down Then keyB_SPC := (keyB_SPC And (Not 4)) Else keyB_SPC := (keyB_SPC Or 4);
      78: //N
        If down Then keyB_SPC := (keyB_SPC And (Not 8)) Else keyB_SPC := (keyB_SPC Or 8);
      79: //O
        If down Then keyY_P := (keyY_P And (Not 2)) Else keyY_P := (keyY_P Or 2);
      80: //P
        If down Then keyY_P := (keyY_P And (Not 1)) Else keyY_P := (keyY_P Or 1);
      81: //Q
        If down Then keyQ_T := (keyQ_T And (Not 1)) Else keyQ_T := (keyQ_T Or 1);
      82: //R
        If down Then keyQ_T := (keyQ_T And (Not 8)) Else keyQ_T := (keyQ_T Or 8);
      83: //S
        If down Then keyA_G := (keyA_G And (Not 2)) Else keyA_G := (keyA_G Or 2);
      84: //T
        If down Then keyQ_T := (keyQ_T And (Not 16)) Else keyQ_T := (keyQ_T Or 16);
      85: //U
        If down Then keyY_P := (keyY_P And (Not 8)) Else keyY_P := (keyY_P Or 8);
      86: //V
        If down Then keyCAPS_V := (keyCAPS_V And (Not 16)) Else keyCAPS_V := (keyCAPS_V Or 16);
      87: //W
        If down Then keyQ_T := (keyQ_T And (Not 2)) Else keyQ_T := (keyQ_T Or 2);
      88: //X
        If down Then keyCAPS_V := (keyCAPS_V And (Not 4)) Else keyCAPS_V := (keyCAPS_V Or 4);
      89: //Y
        If down Then keyY_P := (keyY_P And (Not 16)) Else keyY_P := (keyY_P Or 16);
      90: //Z
        If down Then keyCAPS_V := (keyCAPS_V And (Not 2)) Else keyCAPS_V := (keyCAPS_V Or 2);
      48: //0
        If down Then key6_0 := (key6_0 And (Not 1)) Else key6_0 := (key6_0 Or 1);
      49: //1
        If down Then key1_5 := (key1_5 And (Not 1)) Else key1_5 := (key1_5 Or 1);
      50: //2
        If down Then key1_5 := (key1_5 And (Not 2)) Else key1_5 := (key1_5 Or 2);
      51: //3
        If down Then key1_5 := (key1_5 And (Not 4)) Else key1_5 := (key1_5 Or 4);
      52: //4
        If down Then key1_5 := (key1_5 And (Not 8)) Else key1_5 := (key1_5 Or 8);
      53: //5
        If down Then key1_5 := (key1_5 And (Not 16)) Else key1_5 := (key1_5 Or 16);
      54: //6
        If down Then key6_0 := (key6_0 And (Not 16)) Else key6_0 := (key6_0 Or 16);
      55: //7
        If down Then key6_0 := (key6_0 And (Not 8)) Else key6_0 := (key6_0 Or 8);
      56: //8
        If down Then key6_0 := (key6_0 And (Not 4)) Else key6_0 := (key6_0 Or 4);
      57: //9
        If down Then key6_0 := (key6_0 And (Not 2)) Else key6_0 := (key6_0 Or 2);
      96: //Keypad 0
        If down Then key6_0 := (key6_0 And (Not 1)) Else key6_0 := (key6_0 Or 1);
      97: //Keypad 1
        If down Then key1_5 := (key1_5 And (Not 1)) Else key1_5 := (key1_5 Or 1);
      98: //Keypad 2
        If down Then key1_5 := (key1_5 And (Not 2)) Else key1_5 := (key1_5 Or 2);
      99: //Keypad 3
        If down Then key1_5 := (key1_5 And (Not 4)) Else key1_5 := (key1_5 Or 4);
      100:  //Keypad 4
        If down Then key1_5 := (key1_5 And (Not 8)) Else key1_5 := (key1_5 Or 8);
      101:  //Keypad 5
        If down Then key1_5 := (key1_5 And (Not 16)) Else key1_5 := (key1_5 Or 16);
      102:  //Keypad 6
        If down Then key6_0 := (key6_0 And (Not 16)) Else key6_0 := (key6_0 Or 16);
      103:  //Keypad 7
        If down Then key6_0 := (key6_0 And (Not 8)) Else key6_0 := (key6_0 Or 8);
      104:  //Keypad 8
        If down Then key6_0 := (key6_0 And (Not 4)) Else key6_0 := (key6_0 Or 4);
      105:  //Keypad 9
        If down Then key6_0 := (key6_0 And (Not 2)) Else key6_0 := (key6_0 Or 2);
      106:  //Keypad *
        begin
          If down Then
            keyB_SPC := (keyB_SPC And Not (18))
          Else
          begin
            If SYMB Then
                keyB_SPC := (keyB_SPC Or 16)
            Else
                keyB_SPC := (keyB_SPC Or 18);
          End;
        end;
      107:  //Keypad +
        begin
          If down Then
          begin
            keyH_ENT := (keyH_ENT And (Not 4));
            keyB_SPC := (keyB_SPC And (Not 2));
          end
          Else
          begin
            keyH_ENT := (keyH_ENT Or 4);
            If Not SYMB Then
                keyB_SPC := (keyB_SPC Or 2);
          End;
        end;
      109:  //Keypad -
        begin
          If down Then
          begin
            keyH_ENT := (keyH_ENT And (Not 8));
            keyB_SPC := (keyB_SPC And (Not 2));
          end
          Else
          begin
            keyH_ENT := (keyH_ENT Or 8);
            If Not SYMB Then
                keyB_SPC := (keyB_SPC Or 2);
          End;
        end;
      110:  //Keypad .
        begin
          If down Then
            keyB_SPC := (keyB_SPC And (Not 6))
          Else
          begin
            If SYMB Then
                keyB_SPC := (keyB_SPC Or 4)
            Else
                keyB_SPC := (keyB_SPC Or 6);
          End;
        end;
      111:  //Keypad /
        begin
          If down Then
          begin
            keyCAPS_V := (keyCAPS_V And (Not 16));
            keyB_SPC := (keyB_SPC And (Not 2));
          end
          Else
          begin
            keyCAPS_V := (keyCAPS_V Or 16);
            If Not SYMB Then
                keyB_SPC := (keyB_SPC Or 2);
          End;
        end;
      37: //Left
        begin
          If down Then
          begin
            key1_5 := (key1_5 And (Not 16));
            keyCAPS_V := (keyCAPS_V And (Not 1));
          end
          Else
          begin
            key1_5 := (key1_5 Or 16);
            If Not SYMB Then
                keyB_SPC := (keyB_SPC Or 2);
          End;
        end;
      38: //Up
        begin
          If down Then
          begin
            key6_0 := (key6_0 And (Not 8));
            keyCAPS_V := (keyCAPS_V And (Not 1));
          end
          Else
          begin
            key6_0 := (key6_0 Or 8);
            If Not CAPS Then
                keyCAPS_V := (keyCAPS_V Or 1);
          End;
        end;
      39: //Right
        begin
          If down Then
          begin
            key6_0 := (key6_0 And (Not 4));
            keyCAPS_V := (keyCAPS_V And (Not 1));
          end
          Else
          begin
            key6_0 := (key6_0 Or 4);
            If Not CAPS Then
                keyCAPS_V := (keyCAPS_V Or 1);
          End;
        end;
      40: //Down
        begin
          If down Then
          begin
            key6_0 := (key6_0 And (Not 16));
            keyCAPS_V := (keyCAPS_V And (Not 1));
          end
          Else
          begin
            key6_0 := (key6_0 Or 16);
            If Not CAPS Then
                keyCAPS_V := (keyCAPS_V Or 1);
          End;
        end;
      13: //RETURN
        If down Then keyH_ENT := (keyH_ENT And (Not 1)) Else keyH_ENT := (keyH_ENT Or 1);
      32: //SPACE BAR
        If down Then keyB_SPC := (keyB_SPC And (Not 1)) Else keyB_SPC := (keyB_SPC Or 1);
      187:  //=/+ key
        begin
          If down Then
          begin
            If CAPS Then
                keyH_ENT := (keyH_ENT And (Not 4))
            Else
                keyH_ENT := (keyH_ENT And (Not 2));

            keyB_SPC := (keyB_SPC And (Not 2));
            keyCAPS_V := (keyCAPS_V Or 1);
          end
          Else
          begin
            keyH_ENT := (keyH_ENT Or 4);
            keyH_ENT := (keyH_ENT Or 2);
            keyB_SPC := (keyB_SPC Or 2);
          End;
        end;
      189:  // -/_ key
        begin
          If down Then
          begin
            If CAPS Then
                key6_0 := (key6_0 And (Not 1))
            Else
                keyH_ENT := (keyH_ENT And (Not 8));

            keyB_SPC := (keyB_SPC And (Not 2));
            keyCAPS_V := (keyCAPS_V Or 1);
          end
          Else
          begin
            key6_0 := (key6_0 Or 1);        // Release the Spectrum's '0' key
            keyH_ENT := (keyH_ENT Or 8);    // Release the Spectrum's 'J' key
            keyB_SPC := (keyB_SPC Or 2);    // Release the Symbol Shift key
          End;
        end;
      186:  // ;/: keys
        begin
          If down Then
          begin
            If CAPS Then
                keyCAPS_V := (keyCAPS_V And (Not 2))
            Else
                keyY_P := (keyY_P And (Not 2));

            keyB_SPC := (keyB_SPC And (Not 2));
            keyCAPS_V := (keyCAPS_V Or 1);
          end
          Else
          begin
            keyCAPS_V := (keyCAPS_V Or 2);
            keyY_P := (keyY_P Or 2);
            keyB_SPC := (keyB_SPC Or 2);
          End;
        end;
    End;
End;


procedure TfrmMainWnd.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
var value:word;
begin
    If ssAlt in Shift Then Exit;
    doKey(True, Key, Shift);
   // Joystick

  with gljoystick do
  begin
   if key=up then stickvalue:=stickvalue or $08;
   if key=down then stickvalue:=stickvalue or $04;
   if key=left then stickvalue:=stickvalue or $02;
   if key=right then stickvalue:=stickvalue or $01;
   if key=fire then stickvalue:=stickvalue or $10;
  end;
end;

procedure TfrmMainWnd.FormKeyUp(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  doKey(False, Key, Shift);
  stickvalue:=0;
end;

procedure TfrmMainWnd.mnuFileResetSpectrumClick(Sender: TObject);
  var ii: integer;
begin
  {$IFDEF USE_SOUND}
  If gbSoundEnabled Then
  begin
      waveOutReset(glphWaveOut);
      For ii := 1 To 12 do
          waveOutWrite(glphWaveOut, @gtWavHdr[ii], sizeof(gtWavHdr[ii]));

      glBufNum := 12;
  End;
  {$ENDIF}
  Z80Reset;
  resetKeyboard;
end;

procedure TfrmMainWnd.ResetSound;
var ii: integer;
begin
  {$IFDEF USE_SOUND}
  If gbSoundEnabled Then
  begin
      waveOutReset(glphWaveOut);
      For ii := 1 To 12 do
          waveOutWrite(glphWaveOut, @gtWavHdr[ii], sizeof(gtWavHdr[ii]));

      glBufNum := 12;
  end;
  {$ENDIF}
end;

procedure TfrmMainWnd.ExecConfig;
var oldmodel:byte;
begin
  z80runstate:=Z80PAUSED;
  oldmodel:=glemulatedmodel;
  frmSetup.Showmodal;
  if glemulatedmodel<>oldmodel then
  begin // New Machine has been choosen
   if glautoreboot then mnuFileResetSpectrumClick(nil);
  end;
  ResetSound;
  z80runstate:=Z80RUNNING;
end;

procedure TfrmMainWnd.mnuOptionsGeneralSettingsClick(Sender: TObject);
begin
  frmSetup.PCOptions.ActivePage:=frmSetup.tsGeneric;
  ExecConfig;
end;

procedure TfrmMainWnd.mnuOptionsDisplayClick(Sender: TObject);
begin
  frmSetup.PCOptions.ActivePage:=frmSetup.tsVideo;
  ExecConfig;
end;

procedure TfrmMainWnd.mnuHelpSpectrumKeyboardClick(Sender: TObject);
begin
  //Keyboard box is modeless
  frmKeyboard.Show;
end;

procedure TfrmMainWnd.mnuHelpAboutClick(Sender: TObject);
begin
  z80runstate:=Z80PAUSED;
  frmAbout.ShowModal;
  z80runstate:=Z80RUNNING;
end;

procedure TfrmMainWnd.FileSaveAsDialog;
  var Ext: string;
begin
    dlgSave.execute;

    if dlgSave.FileName <> '' Then
    begin
      Ext := LowerCase(ExtractFileExt(dlgSave.FileName));
      
      if (Ext = '.z80') then
        SaveZ80Snap(dlgSave.FileName)
      else if (Ext = '.rom') then
        SaveROM(dlgSave.FileName)
      else if (Ext = '.sna') then
        SaveSNASnap(dlgSave.FileName)
      else  //save it as a Z80 file
        SaveZ80Snap(dlgSave.FileName);
    end;
end;

procedure TfrmMainWnd.mnuFileSaveAsClick(Sender: TObject);
  var ii:integer;
begin
  {$IFDEF USE_SOUND}
  If gbSoundEnabled Then waveOutReset(glphWaveOut);
  {$ENDIF}

  FileSaveAsDialog;

  {$IFDEF USE_SOUND}
  If gbSoundEnabled Then
  begin
      waveOutReset(glphWaveOut);
      For ii := 1 To 12 do
          waveOutWrite(glphWaveOut, @gtWavHdr[glBufNum], sizeof(gtWavHdr[glBufNum]));

      glBufNum := 12;
  End;
  {$ENDIF}
end;

procedure TfrmMainWnd.FormResize(Sender: TObject);
//var y:byte;
begin
 if (not assigned(display)) or (not visible) then exit;
// y:=Toolbar.Height+Statusbar.Height;
 if (clientwidth>0) then display.Repaint(clientrect)
end;

procedure TfrmMainWnd.FormClose(Sender: TObject; var Action: TCloseAction);
begin
 frmdebug.close;
 onresize:=nil;
 if assigned(Tape) then Tape.Destroy;
 if assigned(Betadisk) then betadisk.destroy;
 if assigned(AY8912) then AY8912.destroy;
 if assigned(Display) then Display.destroy;
end;

procedure TfrmMainWnd.FormActivate(Sender: TObject);
var ThreadHandle: THandle;
begin
 onActivate:=nil;
{$IFDEF USE_DEBUGGING}
  ReportDebugCallCtr:=0;
{$ENDIF}
  Main;
  ThreadHandle := GetCurrentThread;
  SetThreadPriority(ThreadHandle, THREAD_PRIORITY_HIGHEST);
  z80runstate := Z80RUNNING;
  Application.OnIdle:= MyIdleHandler;
  Application.OnActivate:=ContinueEmulation;
  Application.OnDeactivate:=StopEmulation;
  Global_TStates := -glTstatesPerInterrupt;
end;

procedure TfrmMainWnd.mnuNMIClick(Sender: TObject);
begin
 z80_nmi;
end;

procedure TfrmMainWnd.tbResetClick(Sender: TObject);
begin
 Z80Reset;
end;

procedure TfrmMainWnd.tbOpenClick(Sender: TObject);
begin
 mnuFileOpenClick(self);
end;

procedure TfrmMainWnd.tbSaveClick(Sender: TObject);
begin
 mnuFileSaveAsClick(self);
end;

procedure TfrmMainWnd.tbModelClick(Sender: TObject);
begin
 z80runstate:=Z80STOPPED;
 glEmulatedModel:=(sender as ttoolbutton).tag;
 SaveSetting('EmulatedModel', IntToStr(glEmulatedModel));
 Z80Reset;
 z80runstate:=Z80RUNNING;
end;

procedure TfrmMainWnd.tbSetupClick(Sender: TObject);
begin
 mnuOptionsGeneralSettingsClick(self);
end;

procedure TfrmMainWnd.tbKeyboardClick(Sender: TObject);
begin
  frmKeyboard.Show;
end;

procedure TfrmMainWnd.tbPauseClick(Sender: TObject);
begin
 with frmDebug do
  if z80runstate=Z80STOPPED then
   DebugState:=STATE_RUNNING else DebugState:=STATE_PAUSED;
end;

procedure TfrmMainWnd.tbDebugClick(Sender: TObject);
begin
 frmdebug.show;
end;

procedure TfrmMainWnd.mnuInsertDiskClick(Sender: TObject);
begin
 frmSetup.PCOptions.ActivePage:=frmSetup.tsBetadisk;
 ExecConfig;
end;

procedure TfrmMainWnd.FormMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
 try
//  z80runstate:=Z80STOPPED;
  if button=mbRight then SetWindow else
  If (button = mbleft) and (fullwindow) and (windowstate<>wsmaximized) then
  begin
   releasecapture;
   TWincontrol (Self).perform (WM_syscommand, $F012, 0);
  end;
 finally
 // z80runstate:=Z80Running;
 end;
end;

procedure TfrmMainWnd.mnuTapebrowserClick(Sender: TObject);
begin
 with frmSetup.PCOptions do ActivePage:=frmsetup.tsTape;
 ExecConfig;
end;

procedure TfrmMainWnd.mnuDebuggerClick(Sender: TObject);
begin
 frmdebug.show;
end;

procedure TfrmMainWnd.mnuConsoleClick(Sender: TObject);
begin
 frmconsole.show;
end;

procedure TfrmMainWnd.FormPaint(Sender: TObject);
begin
 if (assigned(display)) then
  if clientwidth>0 then display.PaintColorKey(clientrect);
end;

procedure TfrmMainWnd.mnuFullscreenClick(Sender: TObject);
var DM:TMode;
    DC:THandle;
    Bpp,HRes,VRes:Longint;
begin
 ffullscreen:=not ffullscreen;
 if ffullscreen then
  begin // Fullscreen
   if glResolution<>0 then
   begin
    dm:=modes[glResolution-1];
    if (dm.bpp<>0) and (dm.vres<>0) and (dm.hres<>0) then
     changeres(dm.hres,dm.vres,dm.bpp);
   end;
   windowstate:=wsmaximized;
  end else
  begin // Window
   DC   := GetDC(0);
   Bpp  := GetDeviceCaps(DC, BITSPIXEL);
   HRes := GetDeviceCaps(DC, HORZRES);
   VRes := GetDeviceCaps(DC, VERTRES);
   if (DisplayMode.Bpp<>BPP) or (DisplayMode.HRes<>Hres) or
      (DisplayMode.Vres<>Vres) then
       ChangeRes(DisplayMode.hres,DisplayMode.vres,DisplayMode.bpp);
   ReleaseDC(0, DC);
   windowstate:=wsnormal;
  end;
 SetDisplay(glDisplayMode,glRealBorder,glEmulatedFlash,glFrameSkip);
end;

end.


